from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Self

from safeds._utils import _structural_hash
from safeds._validation import _check_bounds, _ClosedBound

if TYPE_CHECKING:
    from safeds.data.image.containers import Image


class ModelImageSize(ABC):
    """
    A container for image size in neural networks.

    Parameters
    ----------
    width:
        the width of the image
    height:
        the height of the image
    channel:
        the channel of the image

    Raises
    ------
    OutOfBoundsError:
        if width or height are below 1
    ValueError
        if an invalid channel is given
    """

    def __init__(self, width: int, height: int, channel: int, *, _ignore_invalid_channel: bool = False) -> None:
        _check_bounds("width", width, lower_bound=_ClosedBound(1))
        _check_bounds("height", height, lower_bound=_ClosedBound(1))
        if not _ignore_invalid_channel and channel not in (1, 3, 4):
            raise ValueError(f"Channel {channel} is not a valid channel option. Use either 1, 3 or 4")
        _check_bounds("channel", channel, lower_bound=_ClosedBound(1))

        self._width = width
        self._height = height
        self._channel = channel

    @classmethod
    def from_image(cls: type[Self], image: Image) -> Self:
        """
        Create a `ModelImageSize` of a given image.

        Parameters
        ----------
        image:
            the given image for the `ModelImageSize`

        Returns
        -------
        image_size:
            the calculated `ModelImageSize`
        """
        return cls(image.width, image.height, image.channel)

    @classmethod
    def from_image_size(cls: type[Self], image_size: ModelImageSize) -> Self:
        """
        Create a `ModelImageSize` of a given image size.

        Parameters
        ----------
        image_size:
            the given image size for the `ModelImageSize`

        Returns
        -------
        image_size:
            the new `ModelImageSize`
        """
        return cls(image_size.width, image_size.height, image_size.channel)

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ModelImageSize):
            return NotImplemented
        return (self is other) or (
            self._width == other._width and self._height == other._height and self._channel == other._channel
        )

    @abstractmethod
    def __hash__(self) -> int:
        return _structural_hash(self.__class__.__name__, self._width, self._height, self._channel)

    def __sizeof__(self) -> int:
        return sys.getsizeof(self._width) + sys.getsizeof(self._height) + sys.getsizeof(self._channel)

    def __str__(self) -> str:
        return f"{self._width}x{self._height}x{self._channel} (WxHxC)"

    @property
    def width(self) -> int:
        """
        Get the width of this `ImageSize` in pixels.

        Returns
        -------
        width:
            The width of this `ImageSize`.
        """
        return self._width

    @property
    def height(self) -> int:
        """
        Get the height of this `ImageSize` in pixels.

        Returns
        -------
        height:
            The height of this `ImageSize`.
        """
        return self._height

    @property
    def channel(self) -> int:
        """
        Get the channel of this `ImageSize` in pixels.

        Returns
        -------
        channel:
            The channel of this `ImageSize`.
        """
        return self._channel


class ConstantImageSize(ModelImageSize):
    """
    A container for constant image size in neural networks.

    Parameters
    ----------
    width:
        the width of the image
    height:
        the height of the image
    channel:
        the channel of the image

    Raises
    ------
    OutOfBoundsError:
        if width or height are below 1
    ValueError
        if an invalid channel is given
    """

    def __eq__(self, other: object) -> bool:
        if isinstance(other, VariableImageSize):
            return other.__eq__(self)
        else:
            return super().__eq__(other)

    def __hash__(self) -> int:
        return super().__hash__()

    def __str__(self) -> str:
        return f"ConstantImageSize | {super().__str__()}"


class VariableImageSize(ModelImageSize):
    """
    A container for variable image size in neural networks.

    With a `VariableImageSize`, all image sizes that are a multiple of `width` and `height` are valid.

    Parameters
    ----------
    width:
        the width of the image
    height:
        the height of the image
    channel:
        the channel of the image

    Raises
    ------
    OutOfBoundsError:
        if width or height are below 1
    ValueError
        if an invalid channel is given
    """

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ModelImageSize):
            return NotImplemented
        return (self is other) or (
            not (self._width % other._width and other._width % self._width)
            and not (self._height % other._height and other._height % self._height)
            and self._channel == other._channel
        )

    def __hash__(self) -> int:
        return super().__hash__()

    def __str__(self) -> str:
        return f"VariableImageSize | {super().__str__()}"



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

from safeds._utils import _structural_hash

from ._cell import Cell

if TYPE_CHECKING:
    import polars as pl

    from ._string_cell import StringCell
    from ._temporal_cell import TemporalCell

T = TypeVar("T")
P = TypeVar("P")
R = TypeVar("R")


class _LazyCell(Cell[T]):
    """
    A single value in a table.

    This implementation only builds an expression that will be evaluated when needed.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, expression: pl.Expr) -> None:
        self._expression: pl.Expr = expression

    # "Boolean" operators (actually bitwise) -----------------------------------

    def __invert__(self) -> Cell[bool]:
        import polars as pl

        return _wrap(self._expression.cast(pl.Boolean).__invert__())

    def __and__(self, other: bool | Cell[bool]) -> Cell[bool]:
        return _wrap(self._expression.__and__(other))

    def __rand__(self, other: bool | Cell[bool]) -> Cell[bool]:
        return _wrap(self._expression.__rand__(other))

    def __or__(self, other: bool | Cell[bool]) -> Cell[bool]:
        return _wrap(self._expression.__or__(other))

    def __ror__(self, other: bool | Cell[bool]) -> Cell[bool]:
        return _wrap(self._expression.__ror__(other))

    def __xor__(self, other: bool | Cell[bool]) -> Cell[bool]:
        return _wrap(self._expression.__xor__(other))

    def __rxor__(self, other: bool | Cell[bool]) -> Cell[bool]:
        return _wrap(self._expression.__rxor__(other))

    # Comparison ---------------------------------------------------------------

    def __eq__(self, other: object) -> Cell[bool]:  # type: ignore[override]
        other = _unwrap(other)
        return _wrap(self._expression.eq_missing(other))

    def __ge__(self, other: Any) -> Cell[bool]:
        other = _unwrap(other)
        return _wrap(self._expression.__ge__(other))

    def __gt__(self, other: Any) -> Cell[bool]:
        other = _unwrap(other)
        return _wrap(self._expression.__gt__(other))

    def __le__(self, other: Any) -> Cell[bool]:
        other = _unwrap(other)
        return _wrap(self._expression.__le__(other))

    def __lt__(self, other: Any) -> Cell[bool]:
        other = _unwrap(other)
        return _wrap(self._expression.__lt__(other))

    def __ne__(self, other: object) -> Cell[bool]:  # type: ignore[override]
        other = _unwrap(other)
        return _wrap(self._expression.ne_missing(other))

    # Numeric operators --------------------------------------------------------

    def __abs__(self) -> Cell[R]:
        return _wrap(self._expression.__abs__())

    def __ceil__(self) -> Cell[R]:
        import polars as pl

        # polars does not yet implement floor for integers
        return _wrap(self._expression.cast(pl.Float64).ceil())

    def __floor__(self) -> Cell[R]:
        import polars as pl

        # polars does not yet implement floor for integers
        return _wrap(self._expression.cast(pl.Float64).floor())

    def __neg__(self) -> Cell[R]:
        return _wrap(self._expression.__neg__())

    def __pos__(self) -> Cell[R]:
        return _wrap(self._expression.__pos__())

    def __add__(self, other: Any) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__add__(other))

    def __radd__(self, other: Any) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__radd__(other))

    def __floordiv__(self, other: Any) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__floordiv__(other))

    def __rfloordiv__(self, other: Any) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__rfloordiv__(other))

    def __mod__(self, other: Any) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__mod__(other))

    def __rmod__(self, other: Any) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__rmod__(other))

    def __mul__(self, other: Any) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__mul__(other))

    def __rmul__(self, other: Any) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__rmul__(other))

    def __pow__(self, other: float | Cell[P]) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__pow__(other))

    def __rpow__(self, other: float | Cell[P]) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__rpow__(other))

    def __sub__(self, other: Any) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__sub__(other))

    def __rsub__(self, other: Any) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__rsub__(other))

    def __truediv__(self, other: Any) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__truediv__(other))

    def __rtruediv__(self, other: Any) -> Cell[R]:
        other = _unwrap(other)
        return _wrap(self._expression.__rtruediv__(other))

    # String representation ----------------------------------------------------

    def __repr__(self) -> str:
        return self._expression.__repr__()

    def __str__(self) -> str:
        return self._expression.__str__()

    # Other --------------------------------------------------------------------

    def __hash__(self) -> int:
        return _structural_hash(self._expression.meta.serialize())

    def __sizeof__(self) -> int:
        return self._expression.__sizeof__()

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def str(self) -> StringCell:
        from ._lazy_string_cell import _LazyStringCell  # circular import

        return _LazyStringCell(self._expression)

    @property
    def dt(self) -> TemporalCell:
        from ._lazy_temporal_cell import _LazyTemporalCell  # circular import

        return _LazyTemporalCell(self._expression)

    # ------------------------------------------------------------------------------------------------------------------
    # Internal
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def _polars_expression(self) -> pl.Expr:
        return self._expression

    def _equals(self, other: object) -> bool:
        if not isinstance(other, _LazyCell):
            return NotImplemented
        if self is other:
            return True
        return self._expression.meta.eq(other._expression.meta)


def _wrap(other: pl.Expr) -> Any:
    return _LazyCell(other)


def _unwrap(other: Any) -> Any:
    if isinstance(other, _LazyCell):
        return other._expression
    return other



from __future__ import annotations

from collections.abc import Collection
from typing import TYPE_CHECKING, TypeVar

from safeds.exceptions import EmptyChoiceError

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any

T = TypeVar("T")


class Choice(Collection[T]):
    """A list of values to choose from in a hyperparameter search."""

    def __init__(self, *args: T) -> None:
        """
        Create a new choice. Duplicate values will be removed.

        Parameters
        ----------
        *args:
            The values to choose from.
        """
        self.elements = list(args)
        if len(args) < 1:
            raise EmptyChoiceError
        self.elements = list(dict.fromkeys(args))

    def __contains__(self, value: Any) -> bool:
        """
        Check if a value is in this choice.

        Parameters
        ----------
        value:
            The value to check.

        Returns
        -------
        is_in_choice:
            Whether the value is in this choice.
        """
        return value in self.elements

    def __iter__(self) -> Iterator[T]:
        """
        Iterate over the values of this choice.

        Returns
        -------
        iterator:
            An iterator over the values of this choice.
        """
        return iter(self.elements)

    def __len__(self) -> int:
        """
        Get the number of values in this choice.

        Returns
        -------
        value_count:
            The number of values in this choice.
        """
        return len(self.elements)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Choice):
            return NotImplemented
        if self is other:
            return True
        return (self is other) or (self.elements == other.elements)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from safeds.data.tabular.containers import Cell


class TemporalCell(ABC):
    """
    Namespace for operations on temporal data.

    This class cannot be instantiated directly. It can only be accessed using the `dt` attribute of a cell.

    Examples
    --------
    >>> from safeds.data.tabular.containers import Column
    >>> import datetime
    >>> column = Column("example", [datetime.date(2022, 1, 9)])
    >>> column.transform(lambda cell: cell.dt.date_to_string("%Y/%m/%d"))
    +------------+
    | example    |
    | ---        |
    | str        |
    +============+
    | 2022/01/09 |
    +------------+
    """

    @abstractmethod
    def century(self) -> Cell[int]:
        """
        Get the century of the underlying date(time) data.

        Returns
        -------
            A cell containing the century as integer.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> import datetime
        >>> column = Column("example", [datetime.date(2022, 1, 1)])
        >>> column.transform(lambda cell: cell.dt.century())
        +---------+
        | example |
        |     --- |
        |     i32 |
        +=========+
        |      21 |
        +---------+
        """

    @abstractmethod
    def weekday(self) -> Cell[int]:
        """
        Get the weekday of the underlying date(time) data.

        Returns
        -------
            A cell containing the weekday as integer.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> import datetime
        >>> column = Column("example", [datetime.date(2022, 1, 1)])
        >>> column.transform(lambda cell: cell.dt.weekday())
        +---------+
        | example |
        |     --- |
        |      i8 |
        +=========+
        |       6 |
        +---------+
        """

    @abstractmethod
    def week(self) -> Cell[int]:
        """
        Get the week of the underlying date(time) data.

        Returns
        -------
            A cell containing the week as integer.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> import datetime
        >>> column = Column("example", [datetime.date(2022, 1, 1)])
        >>> column.transform(lambda cell: cell.dt.week())
        +---------+
        | example |
        |     --- |
        |      i8 |
        +=========+
        |      52 |
        +---------+
        """

    @abstractmethod
    def year(self) -> Cell[int]:
        """
        Get the year of the underlying date(time) data.

        Returns
        -------
            A cell containing the year as integer.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> import datetime
        >>> column = Column("example", [datetime.date(2022, 1, 9)])
        >>> column.transform(lambda cell: cell.dt.year())
        +---------+
        | example |
        |     --- |
        |     i32 |
        +=========+
        |    2022 |
        +---------+
        """

    @abstractmethod
    def month(self) -> Cell[int]:
        """
        Get the month of the underlying date(time) data.

        Returns
        -------
            A cell containing the month as integer.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> import datetime
        >>> column = Column("example", [datetime.date(2022, 1, 9)])
        >>> column.transform(lambda cell: cell.dt.month())
        +---------+
        | example |
        |     --- |
        |      i8 |
        +=========+
        |       1 |
        +---------+
        """

    @abstractmethod
    def day(self) -> Cell[int]:
        """
        Get the day of the underlying date(time) data.

        Returns
        -------
            A cell containing the day as integer.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> import datetime
        >>> column = Column("example", [datetime.date(2022, 1, 9)])
        >>> column.transform(lambda cell: cell.dt.day())
        +---------+
        | example |
        |     --- |
        |      i8 |
        +=========+
        |       9 |
        +---------+
        """

    @abstractmethod
    def datetime_to_string(self, format_string: str = "%Y/%m/%d %H:%M:%S") -> Cell[str]:
        """
        Convert the date value in the cell to a string.

        Parameters
        ----------
        format_string:
            The format string it will be used to convert the data into the string.

        Returns
        -------
        date:
            The string value.

        Raises
        ------
        ValueError
            If the formatstring is invalid.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> import datetime
        >>> column = Column("example", [ datetime.datetime(2022, 1, 9, 23, 29, 1, tzinfo=datetime.UTC)])
        >>> column.transform(lambda cell: cell.dt.datetime_to_string())
        +---------------------+
        | example             |
        | ---                 |
        | str                 |
        +=====================+
        | 2022/01/09 23:29:01 |
        +---------------------+
        """

    @abstractmethod
    def date_to_string(self, format_string: str = "%F") -> Cell[str]:
        """
        Convert the date value in the cell to a string.

        Parameters
        ----------
        format_string:
            The format string it will be used to convert the data into the string.

        Returns
        -------
        date:
            The string value.


        ValueError
            If the formatstring is invalid.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> import datetime
        >>> column = Column("example", [datetime.date(2022, 1, 9)])
        >>> column.transform(lambda cell: cell.dt.date_to_string())
        +------------+
        | example    |
        | ---        |
        | str        |
        +============+
        | 2022-01-09 |
        +------------+
        """

    # ------------------------------------------------------------------------------------------------------------------
    # Internal
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def _equals(self, other: object) -> bool:
        """
        Check if this cell is equal to another object.

        This method is needed because the `__eq__` method is used for element-wise comparisons.
        """



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _get_random_seed, _structural_hash
from safeds.ml.classical._bases import _SupportVectorMachineBase
from safeds.ml.classical.classification import Classifier

if TYPE_CHECKING:
    from sklearn.base import ClassifierMixin


class SupportVectorClassifier(Classifier, _SupportVectorMachineBase):
    """
    Support vector machine for classification.

    Parameters
    ----------
    c:
        The strength of regularization. Must be greater than 0.
    kernel:
        The type of kernel to be used. Defaults to a radial basis function kernel.

    Raises
    ------
    OutOfBoundsError
        If `c` is less than or equal to 0.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(
        self,
        *,
        c: float = 1.0,
        kernel: SupportVectorClassifier.Kernel | None = None,
    ) -> None:
        # Initialize superclasses
        Classifier.__init__(self)
        _SupportVectorMachineBase.__init__(
            self,
            c=c,
            kernel=kernel,
        )

    def __hash__(self) -> int:
        return _structural_hash(
            Classifier.__hash__(self),
            _SupportVectorMachineBase.__hash__(self),
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def kernel(self) -> SupportVectorClassifier.Kernel:
        """The type of kernel used."""
        return self._kernel

    # ------------------------------------------------------------------------------------------------------------------
    # Template methods
    # ------------------------------------------------------------------------------------------------------------------

    def _clone(self) -> SupportVectorClassifier:
        return SupportVectorClassifier(
            c=self._c,
            kernel=self._kernel,
        )

    def _get_sklearn_model(self) -> ClassifierMixin:
        from sklearn.svm import SVC as SklearnSVC  # noqa: N811

        result = SklearnSVC(
            C=self._c,
            random_state=_get_random_seed(),
        )
        self._kernel._apply(result)
        return result



"""Types used to define the schema of a tabular dataset."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._data_type import DataType
    from ._schema import Schema

apipkg.initpkg(
    __name__,
    {
        "DataType": "._data_type:DataType",
        "Schema": "._schema:Schema",
    },
)

__all__ = [
    "DataType",
    "Schema",
]



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Literal

from safeds._utils import _figure_to_image
from safeds._validation import _check_bounds, _check_columns_exist, _ClosedBound
from safeds._validation._check_columns_are_numeric import _check_columns_are_numeric
from safeds.exceptions import ColumnTypeError, NonNumericColumnError

if TYPE_CHECKING:
    from typing import Literal

    from safeds.data.image.containers import Image
    from safeds.data.tabular.containers import Table


class TablePlotter:
    """
    A class that contains plotting methods for a table.

    Parameters
    ----------
    table:
        The table to plot.

    Examples
    --------
    >>> from safeds.data.tabular.containers import Table
    >>> table = Table({"test": [1, 2, 3]})
    >>> plotter = table.plot
    """

    def __init__(self, table: Table):
        self._table: Table = table

    def box_plots(self, *, theme: Literal["dark", "light"] = "light") -> Image:
        """
        Create a box plot for every numerical column.

        Parameters
        ----------
        theme:
            The color theme of the plot. Default is "light".

        Returns
        -------
        plot:
            The box plot(s) as an image.

        Raises
        ------
        NonNumericColumnError
            If the table contains only non-numerical columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a":[1, 2], "b": [3, 42]})
        >>> image = table.plot.box_plots()
        """
        numerical_table = self._table.remove_non_numeric_columns()
        if numerical_table.column_count == 0:
            raise NonNumericColumnError("This table contains only non-numerical columns.")
        from math import ceil

        import matplotlib.pyplot as plt

        style = "dark_background" if theme == "dark" else "default"
        with plt.style.context(style):
            if theme == "dark":
                plt.rcParams.update(
                    {
                        "text.color": "white",
                        "axes.labelcolor": "white",
                        "axes.edgecolor": "white",
                        "xtick.color": "white",
                        "ytick.color": "white",
                    },
                )
            columns = numerical_table.to_columns()
            columns = [column._series.drop_nulls() for column in columns]
            max_width = 3
            number_of_columns = len(columns) if len(columns) <= max_width else max_width
            number_of_rows = ceil(len(columns) / number_of_columns)

            fig, axs = plt.subplots(nrows=number_of_rows, ncols=number_of_columns)
            line = 0
            for i, column in enumerate(columns):
                if i % number_of_columns == 0 and i != 0:
                    line += 1

                if number_of_columns == 1:
                    axs.boxplot(
                        column,
                        patch_artist=True,
                        labels=[numerical_table.column_names[i]],
                    )
                    break

                if number_of_rows == 1:
                    axs[i].boxplot(
                        column,
                        patch_artist=True,
                        labels=[numerical_table.column_names[i]],
                    )

                else:
                    axs[line, i % number_of_columns].boxplot(
                        column,
                        patch_artist=True,
                        labels=[numerical_table.column_names[i]],
                    )

            # removes unused ax indices, so there wont be empty plots
            last_filled_ax_index = len(columns) % number_of_columns
            for i in range(last_filled_ax_index, number_of_columns):
                if number_of_rows != 1 and last_filled_ax_index != 0:
                    fig.delaxes(axs[number_of_rows - 1, i])

            fig.tight_layout()
            return _figure_to_image(fig)

    def violin_plots(self, *, theme: Literal["dark", "light"] = "light") -> Image:
        """
        Create a violin plot for every numerical column.

        Parameters
        ----------
        theme:
            The color theme of the plot. Default is "light".

        Returns
        -------
        plot:
            The violin plot(s) as an image.

        Raises
        ------
        NonNumericColumnError
            If the table contains only non-numerical columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2], "b": [3, 42]})
        >>> image = table.plot.violin_plots()
        """
        numerical_table = self._table.remove_non_numeric_columns()
        if numerical_table.column_count == 0:
            raise NonNumericColumnError("This table contains only non-numerical columns.")
        from math import ceil

        import matplotlib.pyplot as plt

        style = "dark_background" if theme == "dark" else "default"
        with plt.style.context(style):
            if theme == "dark":
                plt.rcParams.update(
                    {
                        "text.color": "white",
                        "axes.labelcolor": "white",
                        "axes.edgecolor": "white",
                        "xtick.color": "white",
                        "ytick.color": "white",
                        "grid.color": "gray",
                        "grid.linewidth": 0.5,
                    },
                )
            else:
                plt.rcParams.update(
                    {
                        "grid.linewidth": 0.5,
                    },
                )

            columns = numerical_table.to_columns()
            columns = [column._series.drop_nulls() for column in columns]
            max_width = 3
            number_of_columns = len(columns) if len(columns) <= max_width else max_width
            number_of_rows = ceil(len(columns) / number_of_columns)

            fig, axs = plt.subplots(nrows=number_of_rows, ncols=number_of_columns)
            line = 0
            for i, column in enumerate(columns):
                data = column.to_list()

                if i % number_of_columns == 0 and i != 0:
                    line += 1

                if number_of_columns == 1:
                    axs.violinplot(
                        data,
                    )
                    axs.set_title(numerical_table.column_names[i])
                    break

                if number_of_rows == 1:
                    axs[i].violinplot(
                        data,
                    )
                    axs[i].set_title(numerical_table.column_names[i])

                else:
                    axs[line, i % number_of_columns].violinplot(
                        data,
                    )
                    axs[line, i % number_of_columns].set_title(numerical_table.column_names[i])

            # removes unused ax indices, so there wont be empty plots
            last_filled_ax_index = len(columns) % number_of_columns
            for i in range(last_filled_ax_index, number_of_columns):
                if number_of_rows != 1 and last_filled_ax_index != 0:
                    fig.delaxes(axs[number_of_rows - 1, i])

            fig.tight_layout()
            return _figure_to_image(fig)

    def correlation_heatmap(self, *, theme: Literal["dark", "light"] = "light") -> Image:
        """
        Plot a correlation heatmap for all numerical columns of this `Table`.

        Parameters
        ----------
        theme:
            The color theme of the plot. Default is "light".

        Returns
        -------
        plot:
            The plot as an image.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"temperature": [10, 15, 20, 25, 30], "sales": [54, 74, 90, 206, 210]})
        >>> image = table.plot.correlation_heatmap()
        """
        import matplotlib.pyplot as plt
        import numpy as np

        style = "dark_background" if theme == "dark" else "default"
        with plt.style.context(style):
            if theme == "dark":
                plt.rcParams.update(
                    {
                        "text.color": "white",
                        "axes.labelcolor": "white",
                        "axes.edgecolor": "white",
                        "xtick.color": "white",
                        "ytick.color": "white",
                    },
                )

            only_numerical = self._table.remove_non_numeric_columns()._data_frame.fill_null(0)

            if self._table.row_count == 0:
                warnings.warn(
                    "An empty table has been used. A correlation heatmap on an empty table will show nothing.",
                    stacklevel=2,
                )

            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    message=(
                        "Attempting to set identical low and high (xlims|ylims) makes transformation singular;"
                        " automatically expanding."
                    ),
                )

                fig, ax = plt.subplots()
                heatmap = plt.imshow(
                    only_numerical.corr().to_numpy(),
                    vmin=-1,
                    vmax=1,
                    cmap="coolwarm",
                )
                ax.set_xticks(
                    np.arange(len(only_numerical.columns)),
                    rotation="vertical",
                    labels=only_numerical.columns,
                )
                ax.set_yticks(np.arange(len(only_numerical.columns)), labels=only_numerical.columns)
                fig.colorbar(heatmap)

                plt.tight_layout()

            return _figure_to_image(fig)

    def histograms(self, *, max_bin_count: int = 10, theme: Literal["dark", "light"] = "light") -> Image:
        """
        Plot a histogram for every column.

        Parameters
        ----------
        max_bin_count:
            The maximum number of bins to use in the histogram. Default is 10.
        theme:
            The color theme of the plot. Default is "light".

        Returns
        -------
        plot:
            The plot as an image.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [2, 3, 5, 1], "b": [54, 74, 90, 2014]})
        >>> image = table.plot.histograms()
        """
        import matplotlib.pyplot as plt

        style = "dark_background" if theme == "dark" else "default"
        with plt.style.context(style):
            if theme == "dark":
                plt.rcParams.update(
                    {
                        "text.color": "white",
                        "axes.labelcolor": "white",
                        "axes.edgecolor": "white",
                        "xtick.color": "white",
                        "ytick.color": "white",
                    },
                )
            import polars as pl

            n_cols = min(3, self._table.column_count)
            n_rows = 1 + (self._table.column_count - 1) // n_cols

            if n_cols == 1 and n_rows == 1:
                fig, axs = plt.subplots(1, 1, tight_layout=True)
                one_col = True
            else:
                fig, axs = plt.subplots(n_rows, n_cols, tight_layout=True, figsize=(n_cols * 3, n_rows * 3))
                one_col = False

            col_names = self._table.column_names
            for col_name, ax in zip(col_names, axs.flatten() if not one_col else [axs], strict=False):
                column = self._table.get_column(col_name)
                distinct_values = column.get_distinct_values()

                ax.set_title(col_name)
                ax.set_xlabel("")
                ax.set_ylabel("")

                if column.is_numeric and len(distinct_values) > max_bin_count:
                    min_val = (column.min() or 0) - 1e-6  # Otherwise the minimum value is not included in the first bin
                    max_val = column.max() or 0
                    bin_count = min(max_bin_count, len(distinct_values))
                    bins = [
                        *(pl.Series(range(bin_count + 1)) / bin_count * (max_val - min_val) + min_val),
                    ]

                    bars = [f"{round((bins[i] + bins[i + 1]) / 2, 2)}" for i in range(len(bins) - 1)]
                    hist = column._series.hist(bins=bins).slice(1, length=max_bin_count).get_column("count").to_numpy()

                    ax.bar(bars, hist, edgecolor="black")
                    ax.set_xticks(range(len(hist)), bars, rotation=45, horizontalalignment="right")
                else:
                    value_counts = (
                        column._series.drop_nulls().value_counts().sort(column.name).slice(0, length=max_bin_count)
                    )
                    distinct_values = value_counts.get_column(column.name).cast(pl.String).to_numpy()
                    hist = value_counts.get_column("count").to_numpy()
                    ax.bar(distinct_values, hist, edgecolor="black")
                    ax.set_xticks(
                        range(len(distinct_values)),
                        distinct_values,
                        rotation=45,
                        horizontalalignment="right",
                    )

            for i in range(len(col_names), n_rows * n_cols):
                fig.delaxes(axs.flatten()[i])  # Remove empty subplots

            return _figure_to_image(fig)

    def line_plot(
        self,
        x_name: str,
        y_names: list[str],
        *,
        show_confidence_interval: bool = True,
        theme: Literal["dark", "light"] = "light",
    ) -> Image:
        """
        Create a line plot for two columns in the table.

        Parameters
        ----------
        x_name:
            The name of the column to be plotted on the x-axis.
        y_names:
            The name(s) of the column(s) to be plotted on the y-axis.
        show_confidence_interval:
            If the confidence interval is shown, per default True.
        theme:
            The color theme of the plot. Default is "light".

        Returns
        -------
        plot:
            The plot as an image.

        Raises
        ------
        ColumnNotFoundError
            If a column does not exist.
        TypeError
            If a column is not numeric.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table(
        ...     {
        ...         "a": [1, 2, 3, 4, 5],
        ...         "b": [2, 3, 4, 5, 6],
        ...     }
        ... )
        >>> image = table.plot.line_plot("a", ["b"])
        """
        _plot_validation(self._table, x_name, y_names)

        import matplotlib.pyplot as plt

        style = "dark_background" if theme == "dark" else "default"
        with plt.style.context(style):
            if theme == "dark":
                plt.rcParams.update(
                    {
                        "text.color": "white",
                        "axes.labelcolor": "white",
                        "axes.edgecolor": "white",
                        "xtick.color": "white",
                        "ytick.color": "white",
                    },
                )
            import polars as pl

            agg_list = []
            for name in y_names:
                agg_list.append(pl.col(name).mean().alias(f"{name}_mean"))
                agg_list.append(pl.count(name).alias(f"{name}_count"))
                agg_list.append(pl.std(name, ddof=0).alias(f"{name}_std"))
            grouped = self._table._lazy_frame.sort(x_name).group_by(x_name).agg(agg_list).collect()

            x = grouped.get_column(x_name)
            y_s = []
            confidence_intervals = []
            for name in y_names:
                y_s.append(grouped.get_column(name + "_mean"))
                confidence_intervals.append(
                    1.96 * grouped.get_column(name + "_std") / grouped.get_column(name + "_count").sqrt(),
                )

            fig, ax = plt.subplots()
            for name, y in zip(y_names, y_s, strict=False):
                ax.plot(x, y, label=name)

            if show_confidence_interval:
                for y, conf in zip(y_s, confidence_intervals, strict=False):
                    ax.fill_between(
                        x,
                        y - conf,
                        y + conf,
                        color="lightblue",
                        alpha=0.15,
                    )
            if len(y_names) > 1:
                name = "values"
            else:
                name = y_names[0]
            ax.set(
                xlabel=x_name,
                ylabel=name,
            )
            ax.legend()
            ax.set_xticks(ax.get_xticks())
            ax.set_xticklabels(
                ax.get_xticklabels(),
                rotation=45,
                horizontalalignment="right",
            )  # rotate the labels of the x Axis to prevent the chance of overlapping of the labels
            fig.tight_layout()

            return _figure_to_image(fig)

    def scatter_plot(self, x_name: str, y_names: list[str], *, theme: Literal["dark", "light"] = "light") -> Image:
        """
        Create a scatter plot for two columns in the table.

        Parameters
        ----------
        x_name:
            The name of the column to be plotted on the x-axis.
        y_names:
            The name(s) of the column(s) to be plotted on the y-axis.
        theme:
            The color theme of the plot. Default is "light".

        Returns
        -------
        plot:
            The plot as an image.

        Raises
        ------
        ColumnNotFoundError
            If a column does not exist.
        TypeError
            If a column is not numeric.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table(
        ...     {
        ...         "a": [1, 2, 3, 4, 5],
        ...         "b": [2, 3, 4, 5, 6],
        ...     }
        ... )
        >>> image = table.plot.scatter_plot("a", ["b"])
        """
        _plot_validation(self._table, x_name, y_names)

        import matplotlib.pyplot as plt

        style = "dark_background" if theme == "dark" else "default"
        with plt.style.context(style):
            if theme == "dark":
                plt.rcParams.update(
                    {
                        "text.color": "white",
                        "axes.labelcolor": "white",
                        "axes.edgecolor": "white",
                        "xtick.color": "white",
                        "ytick.color": "white",
                    },
                )
            fig, ax = plt.subplots()
            for y_name in y_names:
                ax.scatter(
                    x=self._table.get_column(x_name)._series,
                    y=self._table.get_column(y_name)._series,
                    s=64,  # marker size
                    linewidth=1,
                    edgecolor="white",
                    label=y_name,
                )
            if len(y_names) > 1:
                name = "values"
            else:
                name = y_names[0]
            ax.set(
                xlabel=x_name,
                ylabel=name,
            )
            ax.legend()
            ax.set_xticks(ax.get_xticks())
            ax.set_xticklabels(
                ax.get_xticklabels(),
                rotation=45,
                horizontalalignment="right",
            )  # rotate the labels of the x Axis to prevent the chance of overlapping of the labels
            fig.tight_layout()

            return _figure_to_image(fig)

    def moving_average_plot(
        self,
        x_name: str,
        y_name: str,
        window_size: int,
        *,
        theme: Literal["dark", "light"] = "light",
    ) -> Image:
        """
        Create a moving average plot for the y column and plot it by the x column in the table.

        Parameters
        ----------
        x_name:
            The name of the column to be plotted on the x-axis.
        y_name:
            The name of the column to be plotted on the y-axis.
        window_size:
            The size of the moving average window
        theme:
            The color theme of the plot. Default is "light".

        Returns
        -------
        plot:
            The plot as an image.

        Raises
        ------
        ColumnNotFoundError
            If a column does not exist.
        TypeError
            If a column is not numeric.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table(
        ...     {
        ...         "a": [1, 2, 3, 4, 5],
        ...         "b": [2, 3, 4, 5, 6],
        ...     }
        ... )
        >>> image = table.plot.moving_average_plot("a", "b", window_size = 2)
        """
        import matplotlib.pyplot as plt

        style = "dark_background" if theme == "dark" else "default"
        with plt.style.context(style):
            if theme == "dark":
                plt.rcParams.update(
                    {
                        "text.color": "white",
                        "axes.labelcolor": "white",
                        "axes.edgecolor": "white",
                        "xtick.color": "white",
                        "ytick.color": "white",
                    },
                )
            import numpy as np
            import polars as pl

            _plot_validation(self._table, x_name, [y_name])
            for name in [x_name, y_name]:
                if self._table.get_column(name).missing_value_count() >= 1:
                    raise ValueError(
                        f"there are missing values in column '{name}', use transformation to fill missing values "
                        f"or drop the missing values. For a moving average no missing values are allowed.",
                    )

            # Calculate the moving average
            mean_col = pl.col(y_name).mean().alias(y_name)
            grouped = self._table._lazy_frame.sort(x_name).group_by(x_name).agg(mean_col).collect()
            data = grouped
            moving_average = data.select([pl.col(y_name).rolling_mean(window_size).alias("moving_average")])
            # set up the arrays for plotting
            y_data_with_nan = moving_average["moving_average"].to_numpy()
            nan_mask = ~np.isnan(y_data_with_nan)
            y_data = y_data_with_nan[nan_mask]
            x_data = data[x_name].to_numpy()[nan_mask]
            fig, ax = plt.subplots()
            ax.plot(x_data, y_data, label="moving average")
            ax.set(
                xlabel=x_name,
                ylabel=y_name,
            )
            ax.legend()
            if self._table.get_column(x_name).is_temporal and self._table.get_column(x_name).row_count < 9:
                ax.set_xticks(x_data)  # Set x-ticks to the x data points
            ax.set_xticks(ax.get_xticks())
            ax.set_xticklabels(
                ax.get_xticklabels(),
                rotation=45,
                horizontalalignment="right",
            )  # rotate the labels of the x Axis to prevent the chance of overlapping of the labels
            fig.tight_layout()

            return _figure_to_image(fig)

    def histogram_2d(
        self,
        x_name: str,
        y_name: str,
        *,
        x_max_bin_count: int = 10,
        y_max_bin_count: int = 10,
        theme: Literal["dark", "light"] = "light",
    ) -> Image:
        """
        Create a 2D histogram for two columns in the table.

        Parameters
        ----------
        x_name:
            The name of the column to be plotted on the x-axis.
        y_name:
            The name of the column to be plotted on the y-axis.
        x_max_bin_count:
            The maximum number of bins to use in the histogram for the x-axis. Default is 10.
        y_max_bin_count:
            The maximum number of bins to use in the histogram for the y-axis. Default is 10.
        theme:
            The color theme of the plot. Default is "light".

        Returns
        -------
        plot:
            The plot as an image.

        Raises
        ------
        ColumnNotFoundError
            If a column does not exist.
        OutOfBoundsError:
            If x_max_bin_count or y_max_bin_count is less than 1.
        TypeError
            If a column is not numeric.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table(
        ...     {
        ...         "a": [1, 2, 3, 4, 5],
        ...         "b": [2, 3, 4, 5, 6],
        ...     }
        ... )
        >>> image = table.plot.histogram_2d("a", "b")
        """
        _check_bounds("x_max_bin_count", x_max_bin_count, lower_bound=_ClosedBound(1))
        _check_bounds("y_max_bin_count", y_max_bin_count, lower_bound=_ClosedBound(1))
        _plot_validation(self._table, x_name, [y_name])

        import matplotlib.pyplot as plt

        if theme == "dark":
            context = "dark_background"
        else:
            context = "default"

        with plt.style.context(context):
            fig, ax = plt.subplots()

            ax.hist2d(
                x=self._table.get_column(x_name)._series,
                y=self._table.get_column(y_name)._series,
                bins=(x_max_bin_count, y_max_bin_count),
            )
            ax.set_xlabel(x_name)
            ax.set_ylabel(y_name)
            ax.tick_params(
                axis="x",
                labelrotation=45,
            )

            fig.tight_layout()

            return _figure_to_image(fig)


def _plot_validation(table: Table, x_name: str, y_names: list[str]) -> None:
    y_names.append(x_name)
    _check_columns_exist(table, y_names)
    y_names.remove(x_name)
    _check_columns_are_numeric(table, y_names)

    if not table.get_column(x_name).is_numeric and not table.get_column(x_name).is_temporal:
        raise ColumnTypeError(x_name)



"""Utilities for Safe-DS."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._hashing import _structural_hash
    from ._plotting import _figure_to_image
    from ._random import _get_random_seed

apipkg.initpkg(
    __name__,
    {
        "_structural_hash": "._hashing:_structural_hash",
        "_figure_to_image": "._plotting:_figure_to_image",
        "_get_random_seed": "._random:_get_random_seed",
    },
)

__all__ = [
    "_structural_hash",
    "_figure_to_image",
    "_get_random_seed",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _structural_hash

from ._lazy_cell import _LazyCell
from ._temporal_cell import TemporalCell

if TYPE_CHECKING:
    import polars as pl

    from ._cell import Cell


class _LazyTemporalCell(TemporalCell):
    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, expression: pl.Expr) -> None:
        self._expression: pl.Expr = expression

    def __hash__(self) -> int:
        return _structural_hash(self._expression.meta.serialize())

    def __sizeof__(self) -> int:
        return self._expression.__sizeof__()

    # ------------------------------------------------------------------------------------------------------------------
    # Temporal operations
    # ------------------------------------------------------------------------------------------------------------------

    def century(self) -> Cell[int]:
        return _LazyCell(self._expression.dt.century())

    def weekday(self) -> Cell[int]:
        return _LazyCell(self._expression.dt.weekday())

    def week(self) -> Cell[int]:
        return _LazyCell(self._expression.dt.week())

    def year(self) -> Cell[int]:
        return _LazyCell(self._expression.dt.year())

    def month(self) -> Cell[int]:
        return _LazyCell(self._expression.dt.month())

    def day(self) -> Cell[int]:
        return _LazyCell(self._expression.dt.day())

    def datetime_to_string(self, format_string: str = "%Y/%m/%d %H:%M:%S") -> Cell[str]:
        if not _check_format_string(format_string):
            raise ValueError("Invalid format string")
        return _LazyCell(self._expression.dt.to_string(format=format_string))

    def date_to_string(self, format_string: str = "%F") -> Cell[str]:
        if not _check_format_string(format_string):
            # Fehler in _check_format_string
            raise ValueError("Invalid format string")
        return _LazyCell(self._expression.dt.to_string(format=format_string))

    # ------------------------------------------------------------------------------------------------------------------
    # Internal
    # ------------------------------------------------------------------------------------------------------------------

    def _equals(self, other: object) -> bool:
        if not isinstance(other, _LazyTemporalCell):
            return NotImplemented
        if self is other:
            return True
        return self._expression.meta.eq(other._expression.meta)


def _check_format_string(format_string: str) -> bool:
    valid_format_codes = {
        "F": "the standard",
        "a": "abbreviated weekday name",
        "A": "full weekday name",
        "w": "weekday as a decimal number",
        "d": "day of the month as a zero-padded decimal number",
        "b": "abbreviated month name",
        "B": "full month name",
        "m": "month as a zero-padded decimal number",
        "y": "year without century as a zero-padded decimal number",
        "Y": "year with century as a decimal number",
        "H": "hour (24-hour clock) as a zero-padded decimal number",
        "I": "hour (12-hour clock) as a zero-padded decimal number",
        "p": "locale's equivalent of either AM or PM",
        "M": "minute as a zero-padded decimal number",
        "S": "second as a zero-padded decimal number",
        "f": "microsecond as a zero-padded decimal number",
        "z": "UTC offset in the form ±HHMM[SS[.ffffff]]",
        "Z": "time zone name",
        "j": "day of the year as a zero-padded decimal number",
        "U": "week number of the year (Sunday as the first day of the week)",
        "W": "week number of the year (Monday as the first day of the week)",
        "c": "locale's appropriate date and time representation",
        "x": "locale's appropriate date representation",
        "X": "locale's appropriate time representation",
        "%": "a literal '%' character",
    }

    # Keep track of the positions in the string
    i = 0
    n = len(format_string)

    # Iterate over each character in the format string
    while i < n:
        if format_string[i] == "%":
            # Make sure there's at least one character following the '%'
            if i + 1 < n:
                code = format_string[i + 1]
                # Check if the following character is a valid format code
                if code not in valid_format_codes:
                    return False
                i += 2  # Skip ahead past the format code
            else:
                # '%' is at the end of the string with no following format code
                return False
        else:
            i += 1  # Continue to the next character

    return True



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _get_random_seed, _structural_hash
from safeds._validation import _check_bounds, _OpenBound

from ._classifier import Classifier

if TYPE_CHECKING:
    from sklearn.base import ClassifierMixin


class LogisticClassifier(Classifier):
    """
    Regularized logistic regression for classification.

    Parameters
    ----------
    c:
        The regularization strength. Lower values imply stronger regularization. Must be greater than 0.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, *, c: float = 1.0) -> None:
        super().__init__()

        # Validation
        _check_bounds("c", c, lower_bound=_OpenBound(0))

        # Hyperparameters
        self._c: float = c

    def __hash__(self) -> int:
        return _structural_hash(
            super().__hash__(),
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def c(self) -> float:
        """The regularization strength. Lower values imply stronger regularization."""
        return self._c

    # ------------------------------------------------------------------------------------------------------------------
    # Template methods
    # ------------------------------------------------------------------------------------------------------------------

    def _clone(self) -> LogisticClassifier:
        return LogisticClassifier(c=self.c)

    def _get_sklearn_model(self) -> ClassifierMixin:
        from sklearn.linear_model import LogisticRegression as SklearnLogisticRegression

        return SklearnLogisticRegression(
            random_state=_get_random_seed(),
            n_jobs=-1,
            C=self.c,
        )



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from safeds._utils import _structural_hash
from safeds.data.labeled.containers import TimeSeriesDataset
from safeds.data.tabular.containers import Column, Table

from ._input_converter import InputConversion

if TYPE_CHECKING:
    from torch import Tensor
    from torch.utils.data import DataLoader


class InputConversionTimeSeries(InputConversion[TimeSeriesDataset, Table]):
    """The input conversion for a neural network, defines the input parameters for the neural network."""

    def __init__(
        self,
    ) -> None:
        self._window_size = 0
        self._forecast_horizon = 0
        self._first = True
        self._target_name: str = ""
        self._feature_names: list[str] = []
        self._continuous: bool = False
        self._extra_names: list[str] = []

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self is other) or (
            self._window_size == other._window_size
            and self._forecast_horizon == other._forecast_horizon
            and self._first == other._first
            and self._target_name == other._target_name
            and self._feature_names == other._feature_names
            and self._continuous == other._continuous
            and self._extra_names == other._extra_names
        )

    def __hash__(self) -> int:
        return _structural_hash(
            self.__class__.__name__,
            self._window_size,
            self._forecast_horizon,
            self._target_name,
            self._feature_names,
            self._continuous,
            self._extra_names,
        )

    def __sizeof__(self) -> int:
        return (
            sys.getsizeof(self._window_size)
            + sys.getsizeof(self._forecast_horizon)
            + sys.getsizeof(self._target_name)
            + sys.getsizeof(self._feature_names)
            + sys.getsizeof(self._continuous)
            + sys.getsizeof(self._extra_names)
        )

    @property
    def _data_size(self) -> int:
        """
        Gives the size for the input of an internal layer.

        Returns
        -------
        size:
            The size of the input for the neural network
        """
        return (len(self._feature_names) + 1) * self._window_size

    def _data_conversion_fit(
        self,
        input_data: TimeSeriesDataset,
        batch_size: int,
        num_of_classes: int = 1,
    ) -> DataLoader:
        self._num_of_classes = num_of_classes
        return input_data._into_dataloader_with_window(
            self._window_size,
            self._forecast_horizon,
            batch_size,
            continuous=self._continuous,
        )

    def _data_conversion_predict(self, input_data: Table, batch_size: int) -> DataLoader:
        data: TimeSeriesDataset
        data = input_data.to_time_series_dataset(
            target_name=self._target_name,
            window_size=self._window_size,
            extra_names=self._extra_names,
            forecast_horizon=self._forecast_horizon,
            continuous=self._continuous,
        )
        return data._into_dataloader_with_window_predict(
            self._window_size,
            self._forecast_horizon,
            batch_size,
        )

    def _data_conversion_output(
        self,
        input_data: Table,
        output_data: Tensor,
    ) -> TimeSeriesDataset:
        table_data: Table
        window_size: int = self._window_size
        forecast_horizon: int = self._forecast_horizon
        table_data = input_data
        input_data_table = table_data.slice_rows(start=window_size + forecast_horizon)

        return input_data_table.replace_column(
            self._target_name,
            [Column(self._target_name, output_data.tolist())],
        ).to_time_series_dataset(
            target_name=self._target_name,
            window_size=self._window_size,
            extra_names=self._extra_names,
            forecast_horizon=self._forecast_horizon,
            continuous=self._continuous,
        )

    def _is_fit_data_valid(self, input_data: TimeSeriesDataset) -> bool:
        if self._first:
            self._window_size = input_data.window_size
            self._forecast_horizon = input_data.forecast_horizon
            self._feature_names = input_data.features.column_names
            self._target_name = input_data.target.name
            self._continuous = input_data._continuous
            self._first = False
            self._extra_names = input_data.extras.column_names
        return (
            sorted(input_data.features.column_names).__eq__(
                sorted(self._feature_names),
            )
            and input_data.target.name == self._target_name
        )

    def _is_predict_data_valid(self, input_data: Table) -> bool:
        for name in self._feature_names:
            if name not in input_data.column_names:
                return False
        if self._target_name not in input_data.column_names:
            return False
        return True



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds.exceptions import ColumnTypeError

if TYPE_CHECKING:
    from collections.abc import Container

    from safeds.data.tabular.containers import Column, Table
    from safeds.data.tabular.typing import Schema


def _check_column_is_numeric(
    column: Column,
    *,
    operation: str = "do a numeric operation",
) -> None:
    """
    Check if the column is numeric and raise an error if it is not.

    Parameters
    ----------
    column:
        The column to check.
    operation:
        The operation that is performed on the column. This is used in the error message.

    Raises
    ------
    ColumnTypeError
        If the column is not numeric.
    """
    if not column.type.is_numeric:
        message = _build_error_message([column.name], operation)
        raise ColumnTypeError(message)


def _check_columns_are_numeric(
    table_or_schema: Table | Schema,
    column_names: str | list[str],
    *,
    operation: str = "do a numeric operation",
) -> None:
    """
    Check if the columns with the specified names are numeric and raise an error if they are not.

    Missing columns are ignored. Use `_check_columns_exist` to check for missing columns.

    Parameters
    ----------
    table_or_schema:
        The table or schema to check.
    column_names:
        The column names to check.
    operation:
        The operation that is performed on the columns. This is used in the error message.

    Raises
    ------
    ColumnTypeError
        If a column exists but is not numeric.
    """
    from safeds.data.tabular.containers import Table  # circular import

    if isinstance(table_or_schema, Table):
        table_or_schema = table_or_schema.schema
    if isinstance(column_names, str):
        column_names = [column_names]

    if len(column_names) > 1:
        # Create a set for faster containment checks
        known_names: Container = set(table_or_schema.column_names)
    else:
        known_names = table_or_schema.column_names

    non_numeric_names = [
        name for name in column_names if name in known_names and not table_or_schema.get_column_type(name).is_numeric
    ]
    if non_numeric_names:
        message = _build_error_message(non_numeric_names, operation)
        raise ColumnTypeError(message)


def _build_error_message(non_numeric_names: list[str], operation: str) -> str:
    return f"Tried to {operation} on non-numeric columns {non_numeric_names}."



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _structural_hash
from safeds._validation import _check_bounds, _check_columns_exist, _ClosedBound
from safeds._validation._check_columns_are_numeric import _check_columns_are_numeric
from safeds.data.tabular.containers import Table
from safeds.exceptions import (
    NonNumericColumnError,
    TransformerNotFittedError,
)

from ._table_transformer import TableTransformer

if TYPE_CHECKING:
    from sklearn.preprocessing import KBinsDiscretizer as sk_KBinsDiscretizer


class Discretizer(TableTransformer):
    """
    The Discretizer bins continuous data into intervals.

    Parameters
    ----------
    bin_count:
        The number of bins to be created.
    column_names:
        The list of columns used to fit the transformer. If `None`, all numeric columns are used.

    Raises
    ------
    OutOfBoundsError
        If the given `bin_count` is less than 2.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(
        self,
        bin_count: int = 5,
        *,
        column_names: str | list[str] | None = None,
    ) -> None:
        TableTransformer.__init__(self, column_names)

        _check_bounds("bin_count", bin_count, lower_bound=_ClosedBound(2))

        self._wrapped_transformer: sk_KBinsDiscretizer | None = None
        self._bin_count = bin_count

    def __hash__(self) -> int:
        return _structural_hash(
            TableTransformer.__hash__(self),
            self._bin_count,
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def is_fitted(self) -> bool:
        """Whether the transformer is fitted."""
        return self._wrapped_transformer is not None

    @property
    def bin_count(self) -> int:
        """The number of bins to be created."""
        return self._bin_count

    # ------------------------------------------------------------------------------------------------------------------
    # Learning and transformation
    # ------------------------------------------------------------------------------------------------------------------

    def fit(self, table: Table) -> Discretizer:
        """
        Learn a transformation for a set of columns in a table.

        This transformer is not modified.

        Parameters
        ----------
        table:
            The table used to fit the transformer.

        Returns
        -------
        fitted_transformer:
            The fitted transformer.

        Raises
        ------
        ValueError
            If the table is empty.
        NonNumericColumnError
            If one of the columns, that should be fitted is non-numeric.
        ColumnNotFoundError
            If one of the columns, that should be fitted is not in the table.
        """
        from sklearn.preprocessing import KBinsDiscretizer as sk_KBinsDiscretizer

        if table.row_count == 0:
            raise ValueError("The Discretizer cannot be fitted because the table contains 0 rows")

        if self._column_names is None:
            column_names = [name for name in table.column_names if table.get_column_type(name).is_numeric]
        else:
            column_names = self._column_names
            _check_columns_exist(table, column_names)
            _check_columns_are_numeric(table, column_names, operation="fit a Discretizer")

        wrapped_transformer = sk_KBinsDiscretizer(n_bins=self._bin_count, encode="ordinal")
        wrapped_transformer.set_output(transform="polars")
        wrapped_transformer.fit(
            table.remove_columns_except(column_names)._data_frame,
        )

        result = Discretizer(self._bin_count, column_names=column_names)
        result._wrapped_transformer = wrapped_transformer

        return result

    def transform(self, table: Table) -> Table:
        """
        Apply the learned transformation to a table.

        The table is not modified.

        Parameters
        ----------
        table:
            The table to which the learned transformation is applied.

        Returns
        -------
        transformed_table:
            The transformed table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        ValueError
            If the table is empty.
        ColumnNotFoundError
            If one of the columns, that should be transformed is not in the table.
        NonNumericColumnError
            If one of the columns, that should be fitted is non-numeric.
        """
        # Transformer has not been fitted yet
        if self._wrapped_transformer is None or self._column_names is None:
            raise TransformerNotFittedError

        if table.row_count == 0:
            raise ValueError("The table cannot be transformed because it contains 0 rows")

        # Input table does not contain all columns used to fit the transformer
        _check_columns_exist(table, self._column_names)

        for column in self._column_names:
            if not table.get_column(column).type.is_numeric:
                raise NonNumericColumnError(f"{column} is of type {table.get_column(column).type}.")

        new_data = self._wrapped_transformer.transform(
            table.remove_columns_except(self._column_names)._data_frame,
        )
        return Table._from_polars_lazy_frame(
            table._lazy_frame.update(new_data.lazy()),
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Self

if TYPE_CHECKING:
    from torch import nn

    from safeds.ml.hyperparameters import Choice
    from safeds.ml.nn.typing import ModelImageSize


class Layer(ABC):
    @abstractmethod
    def __init__(self) -> None:
        pass  # pragma: no cover

    @abstractmethod
    def _get_internal_layer(self, **kwargs: Any) -> nn.Module:
        pass  # pragma: no cover

    @property
    @abstractmethod
    def input_size(self) -> int | ModelImageSize:
        pass  # pragma: no cover

    @property
    @abstractmethod
    def output_size(self) -> int | ModelImageSize | Choice[int]:
        pass  # pragma: no cover

    @abstractmethod
    def _set_input_size(self, input_size: int | ModelImageSize) -> None:
        pass  # pragma: no cover

    @abstractmethod
    def _contains_choices(self) -> bool:
        pass  # pragma: no cover

    @abstractmethod
    def _get_layers_for_all_choices(self) -> list[Self]:
        pass  # pragma: no cover

    @abstractmethod
    def __hash__(self) -> int:
        pass  # pragma: no cover

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        pass  # pragma: no cover

    @abstractmethod
    def __sizeof__(self) -> int:
        pass  # pragma: no cover



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds.exceptions import ColumnNotFoundError

if TYPE_CHECKING:
    from collections.abc import Container

    from safeds.data.tabular.containers import Table
    from safeds.data.tabular.typing import Schema


def _check_columns_exist(table_or_schema: Table | Schema, requested_names: str | list[str]) -> None:
    """
    Check if the specified column names exist in the table or schema and raise an error if they do not.

    Parameters
    ----------
    table_or_schema:
        The table or schema to check.
    requested_names:
        The column names to check.

    Raises
    ------
    ColumnNotFoundError
        If a column name does not exist.
    """
    from safeds.data.tabular.containers import Table  # circular import

    if isinstance(table_or_schema, Table):
        table_or_schema = table_or_schema.schema
    if isinstance(requested_names, str):
        requested_names = [requested_names]

    if len(requested_names) > 1:
        # Create a set for faster containment checks
        known_names: Container = set(table_or_schema.column_names)
    else:
        known_names = table_or_schema.column_names

    unknown_names = [name for name in requested_names if name not in known_names]
    if unknown_names:
        message = _build_error_message(table_or_schema, unknown_names)
        raise ColumnNotFoundError(message)


def _build_error_message(schema: Schema, unknown_names: list[str]) -> str:
    message = "Could not find column(s):"

    for unknown_name in unknown_names:
        similar_columns = _get_similar_column_names(schema, unknown_name)
        message += f"\n    - '{unknown_name}'"
        if similar_columns:
            message += f": Did you mean one of {similar_columns}?"

    return message


def _get_similar_column_names(schema: Schema, unknown_name: str) -> list[str]:
    from difflib import get_close_matches

    return get_close_matches(
        unknown_name,
        schema.column_names,
        n=3,
    )



from __future__ import annotations

from abc import ABC, abstractmethod

from safeds.data.labeled.containers import TabularDataset
from safeds.data.tabular.containers import Column, Table
from safeds.exceptions import ColumnLengthMismatchError


class RegressionMetrics(ABC):
    """A collection of regression metrics."""

    @abstractmethod
    def __init__(self) -> None: ...

    @staticmethod
    def summarize(predicted: Column | TabularDataset, expected: Column | TabularDataset) -> Table:
        """
        Summarize regression metrics on the given data.

        Parameters
        ----------
        predicted:
            The predicted target values produced by the regressor.
        expected:
            The expected target values.

        Returns
        -------
        metrics:
            A table containing the regression metrics.
        """
        expected = _extract_target(expected)
        predicted = _extract_target(predicted)
        _check_equal_length(predicted, expected)

        coefficient_of_determination = RegressionMetrics.coefficient_of_determination(expected, predicted)
        mean_absolute_error = RegressionMetrics.mean_absolute_error(expected, predicted)
        mean_squared_error = RegressionMetrics.mean_squared_error(expected, predicted)
        median_absolute_deviation = RegressionMetrics.median_absolute_deviation(expected, predicted)

        return Table(
            {
                "metric": [
                    "coefficient_of_determination",
                    "mean_absolute_error",
                    "mean_squared_error",
                    "median_absolute_deviation",
                ],
                "value": [
                    coefficient_of_determination,
                    mean_absolute_error,
                    mean_squared_error,
                    median_absolute_deviation,
                ],
            },
        )

    @staticmethod
    def coefficient_of_determination(predicted: Column | TabularDataset, expected: Column | TabularDataset) -> float:
        """
        Compute the coefficient of determination (R²) on the given data.

        The coefficient of determination compares the regressor's predictions to another model that always predicts the
        mean of the target values. It is a measure of how well the regressor explains the variance in the target values.

        The **higher** the coefficient of determination, the better the regressor. Results range from negative infinity
        to 1.0. You can interpret the coefficient of determination as follows:

        | R²         | Interpretation                                                                             |
        | ---------- | ------------------------------------------------------------------------------------------ |
        | 1.0        | The model perfectly predicts the target values. Did you overfit?                           |
        | (0.0, 1.0) | The model is better than predicting the mean of the target values. You should be here.     |
        | 0.0        | The model is as good as predicting the mean of the target values. Try something else.      |
        | (-∞, 0.0)  | The model is worse than predicting the mean of the target values. Something is very wrong. |

        **Note:** Some other libraries call this metric `r2_score`.

        Parameters
        ----------
        predicted:
            The predicted target values produced by the regressor.
        expected:
            The expected target values.

        Returns
        -------
        coefficient_of_determination:
            The calculated coefficient of determination.
        """
        expected = _extract_target(expected)
        predicted = _extract_target(predicted)
        _check_equal_length(predicted, expected)

        residual_sum_of_squares = (expected._series - predicted._series).pow(2).sum()
        total_sum_of_squares = (expected._series - expected._series.mean()).pow(2).sum()

        if total_sum_of_squares == 0:
            if residual_sum_of_squares == 0:
                return 1.0  # Everything was predicted correctly
            else:
                return 0.0  # Model could not even predict constant data

        return 1 - residual_sum_of_squares / total_sum_of_squares

    @staticmethod
    def mean_absolute_error(predicted: Column | TabularDataset, expected: Column | TabularDataset) -> float:
        """
        Compute the mean absolute error (MAE) on the given data.

        The mean absolute error is the average of the absolute differences between the predicted and expected target
        values. The **lower** the mean absolute error, the better the regressor. Results range from 0.0 to positive
        infinity.

        Parameters
        ----------
        predicted:
            The predicted target values produced by the regressor.
        expected:
            The expected target values.

        Returns
        -------
        mean_absolute_error:
            The calculated mean absolute error.
        """
        expected = _extract_target(expected)
        predicted = _extract_target(predicted)
        _check_equal_length(predicted, expected)

        if expected.row_count == 0:
            return 0.0  # Everything was predicted correctly (since there is nothing to predict)

        return (expected._series - predicted._series).abs().mean()

    @staticmethod
    def mean_directional_accuracy(predicted: Column | TabularDataset, expected: Column | TabularDataset) -> float:
        """
        Compute the mean directional accuracy (MDA) on the given data.

        This metric compares two consecutive target values and checks if the predicted direction (down/unchanged/up)
        matches the expected direction. The mean directional accuracy is the proportion of correctly predicted
        directions. The **higher** the mean directional accuracy, the better the regressor. Results range from 0.0 to
        1.0.

        This metric is useful for time series data, where the order of the target values has a meaning. It is not useful
        for other types of data. Because of this, it is not included in the `summarize` method.

        Parameters
        ----------
        predicted:
            The predicted target values produced by the regressor.
        expected:
            The expected target values.

        Returns
        -------
        mean_directional_accuracy:
            The calculated mean directional accuracy.
        """
        expected = _extract_target(expected)
        predicted = _extract_target(predicted)
        _check_equal_length(predicted, expected)

        if expected.row_count == 0:
            return 1.0

        # Calculate the differences between the target values
        predicted_directions = predicted._series.diff().sign()
        expected_directions = expected._series.diff().sign()

        return predicted_directions.eq(expected_directions).mean()

    @staticmethod
    def mean_squared_error(predicted: Column | TabularDataset, expected: Column | TabularDataset) -> float:
        """
        Compute the mean squared error (MSE) on the given data.

        The mean squared error is the average of the squared differences between the predicted and expected target
        values. The **lower** the mean squared error, the better the regressor. Results range from 0.0 to positive
        infinity.

        **Note:** To get the root mean squared error (RMSE), take the square root of the result.

        Parameters
        ----------
        predicted:
            The predicted target values produced by the regressor.
        expected:
            The expected target values.

        Returns
        -------
        mean_squared_error:
            The calculated mean squared error.
        """
        expected = _extract_target(expected)
        predicted = _extract_target(predicted)
        _check_equal_length(predicted, expected)

        if expected.row_count == 0:
            return 0.0  # Everything was predicted correctly (since there is nothing to predict)

        return (expected._series - predicted._series).pow(2).mean()

    @staticmethod
    def median_absolute_deviation(predicted: Column | TabularDataset, expected: Column | TabularDataset) -> float:
        """
        Compute the median absolute deviation (MAD) on the given data.

        The median absolute deviation is the median of the absolute differences between the predicted and expected
        target values. The **lower** the median absolute deviation, the better the regressor. Results range from 0.0 to
        positive infinity.

        Parameters
        ----------
        predicted:
            The predicted target values produced by the regressor.
        expected:
            The expected target values.

        Returns
        -------
        median_absolute_deviation:
            The calculated median absolute deviation.
        """
        expected = _extract_target(expected)
        predicted = _extract_target(predicted)
        _check_equal_length(predicted, expected)

        if expected.row_count == 0:
            return 0.0

        return (expected._series - predicted._series).abs().median()


def _extract_target(column_or_dataset: Column | TabularDataset) -> Column:
    """Extract the target column from the given column or dataset."""
    if isinstance(column_or_dataset, TabularDataset):
        return column_or_dataset.target
    else:
        return column_or_dataset


# TODO: collect validation in one place?
def _check_equal_length(column1: Column, column2: Column) -> None:
    """
    Check if the columns have the same length and raise an error if they do not.

    Parameters
    ----------
    column1:
        The first column.
    column2:
        The second column.

    Raises
    ------
    ValueError
        If the columns have different lengths.
    """
    if column1.row_count != column2.row_count:
        ColumnLengthMismatchError("")  # TODO: pass list of columns to exception, let it handle the formatting



"""Classes that can store tabular data."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._cell import Cell
    from ._column import Column
    from ._row import Row
    from ._string_cell import StringCell
    from ._table import Table
    from ._temporal_cell import TemporalCell

apipkg.initpkg(
    __name__,
    {
        "Cell": "._cell:Cell",
        "Column": "._column:Column",
        "Row": "._row:Row",
        "StringCell": "._string_cell:StringCell",
        "TemporalCell": "._temporal_cell:TemporalCell",
        "Table": "._table:Table",
    },
)

__all__ = [
    "Cell",
    "Column",
    "Row",
    "StringCell",
    "TemporalCell",
    "Table",
]



# These class must not be nested inside a function, since pickle cannot serialize local classes. Because of this, the
# slow import of torch must be on the global level. To still evaluate the torch import lazily, the classes are moved to
# a separate file.

from __future__ import annotations

from typing import Literal

from torch import Tensor, nn  # slow import

from safeds._config import _init_default_device


class _InternalConvolutional2DLayer(nn.Module):
    def __init__(
        self,
        input_size: int,
        output_size: int,
        kernel_size: int,
        activation_function: Literal["sigmoid", "relu", "softmax"],
        padding: int,
        stride: int,
        transpose: bool,
        output_padding: int,
    ):
        super().__init__()

        _init_default_device()

        if transpose:
            self._layer = nn.ConvTranspose2d(
                in_channels=input_size,
                out_channels=output_size,
                kernel_size=kernel_size,
                padding=padding,
                stride=stride,
                output_padding=output_padding,
            )
        else:
            self._layer = nn.Conv2d(
                in_channels=input_size,
                out_channels=output_size,
                kernel_size=kernel_size,
                padding=padding,
                stride=stride,
            )
        match activation_function:
            case "sigmoid":
                self._fn = nn.Sigmoid()
            case "relu":
                self._fn = nn.ReLU()
            case "softmax":
                self._fn = nn.Softmax()

    def forward(self, x: Tensor) -> Tensor:
        return self._fn(self._layer(x))


class _InternalDropoutLayer(nn.Module):
    def __init__(self, probability: float) -> None:
        super().__init__()

        _init_default_device()

        self._layer = nn.Dropout(probability)

    def forward(self, x: Tensor) -> Tensor:
        return self._layer(x)


class _InternalFlattenLayer(nn.Module):
    def __init__(self) -> None:
        super().__init__()

        _init_default_device()

        self._layer = nn.Flatten()

    def forward(self, x: Tensor) -> Tensor:
        return self._layer(x)


class _InternalForwardLayer(nn.Module):
    def __init__(self, input_size: int, output_size: int, activation_function: str):
        super().__init__()

        _init_default_device()

        self._layer = nn.Linear(input_size, output_size)
        match activation_function:
            case "sigmoid":
                self._fn = nn.Sigmoid()
            case "relu":
                self._fn = nn.ReLU()
            case "softmax":
                self._fn = nn.Softmax()
            case "none":
                self._fn = None
            case _:
                raise ValueError("Unknown Activation Function: " + activation_function)

    def forward(self, x: Tensor) -> Tensor:
        return self._fn(self._layer(x)) if self._fn is not None else self._layer(x)


class _InternalLSTMLayer(nn.Module):
    def __init__(self, input_size: int, output_size: int, activation_function: str):
        super().__init__()

        _init_default_device()

        self._layer = nn.LSTM(input_size, output_size)
        match activation_function:
            case "sigmoid":
                self._fn = nn.Sigmoid()
            case "relu":
                self._fn = nn.ReLU()
            case "softmax":
                self._fn = nn.Softmax()
            case "none":
                self._fn = None
            case _:
                raise ValueError("Unknown Activation Function: " + activation_function)

    def forward(self, x: Tensor) -> Tensor:
        return self._fn(self._layer(x)[0]) if self._fn is not None else self._layer(x)[0]


class _InternalPooling2DLayer(nn.Module):
    def __init__(self, strategy: Literal["max", "avg"], kernel_size: int, padding: int, stride: int):
        super().__init__()

        _init_default_device()

        match strategy:
            case "max":
                self._layer = nn.MaxPool2d(kernel_size=kernel_size, padding=padding, stride=stride)
            case "avg":
                self._layer = nn.AvgPool2d(kernel_size=kernel_size, padding=padding, stride=stride)

    def forward(self, x: Tensor) -> Tensor:
        return self._layer(x)


class _InternalGRULayer(nn.Module):
    def __init__(self, input_size: int, output_size: int, activation_function: str):
        super().__init__()

        _init_default_device()

        self._layer = nn.GRU(input_size, output_size)
        match activation_function:
            case "sigmoid":
                self._fn = nn.Sigmoid()
            case "relu":
                self._fn = nn.ReLU()
            case "softmax":
                self._fn = nn.Softmax()
            case "none":
                self._fn = None
            case _:
                raise ValueError("Unknown Activation Function: " + activation_function)

    def forward(self, x: Tensor) -> Tensor:
        return self._fn(self._layer(x)[0]) if self._fn is not None else self._layer(x)[0]



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING

from safeds.data.labeled.containers import TabularDataset
from safeds.exceptions import ModelNotFittedError
from safeds.ml.classical import SupervisedModel
from safeds.ml.metrics import ClassificationMetrics

if TYPE_CHECKING:
    from typing import Any

    from safeds.data.tabular.containers import Table


class Classifier(SupervisedModel, ABC):
    """A model for classification tasks."""

    # ------------------------------------------------------------------------------------------------------------------
    # Metrics
    # ------------------------------------------------------------------------------------------------------------------

    def summarize_metrics(
        self,
        validation_or_test_set: Table | TabularDataset,
        positive_class: Any,
    ) -> Table:
        """
        Summarize the classifier's metrics on the given data.

        **Note:** The model must be fitted.

        Parameters
        ----------
        validation_or_test_set:
            The validation or test set.
        positive_class:
            The class to be considered positive. All other classes are considered negative.

        Returns
        -------
        metrics:
            A table containing the classifier's metrics.

        Raises
        ------
        ModelNotFittedError
            If the classifier has not been fitted yet.
        """
        if not self.is_fitted:
            raise ModelNotFittedError

        validation_or_test_set = _extract_table(validation_or_test_set)

        return ClassificationMetrics.summarize(
            self.predict(validation_or_test_set),
            validation_or_test_set.get_column(self.get_target_name()),
            positive_class,
        )

    def accuracy(self, validation_or_test_set: Table | TabularDataset) -> float:
        """
        Compute the accuracy of the classifier on the given data.

        The accuracy is the proportion of predicted target values that were correct. The **higher** the accuracy, the
        better. Results range from 0.0 to 1.0.

        **Note:** The model must be fitted.

        Parameters
        ----------
        validation_or_test_set:
            The validation or test set.

        Returns
        -------
        accuracy:
            The classifier's accuracy.

        Raises
        ------
        ModelNotFittedError
            If the classifier has not been fitted yet.
        """
        if not self.is_fitted:
            raise ModelNotFittedError

        validation_or_test_set = _extract_table(validation_or_test_set)

        return ClassificationMetrics.accuracy(
            self.predict(validation_or_test_set),
            validation_or_test_set.get_column(self.get_target_name()),
        )

    def f1_score(
        self,
        validation_or_test_set: Table | TabularDataset,
        positive_class: Any,
    ) -> float:
        """
        Compute the classifier's F₁ score on the given data.

        The F₁ score is the harmonic mean of precision and recall. The **higher** the F₁ score, the better the
        classifier. Results range from 0.0 to 1.0.

        **Note:** The model must be fitted.

        Parameters
        ----------
        validation_or_test_set:
            The validation or test set.
        positive_class:
            The class to be considered positive. All other classes are considered negative.

        Returns
        -------
        f1_score:
            The classifier's F₁ score.

        Raises
        ------
        ModelNotFittedError
            If the classifier has not been fitted yet.
        """
        if not self.is_fitted:
            raise ModelNotFittedError

        validation_or_test_set = _extract_table(validation_or_test_set)

        return ClassificationMetrics.f1_score(
            self.predict(validation_or_test_set),
            validation_or_test_set.get_column(self.get_target_name()),
            positive_class,
        )

    def precision(
        self,
        validation_or_test_set: Table | TabularDataset,
        positive_class: Any,
    ) -> float:
        """
        Compute the classifier's precision on the given data.

        The precision is the proportion of positive predictions that were correct. The **higher** the precision, the
        better the classifier. Results range from 0.0 to 1.0.

        **Note:** The model must be fitted.

        Parameters
        ----------
        validation_or_test_set:
            The validation or test set.
        positive_class:
            The class to be considered positive. All other classes are considered negative.

        Returns
        -------
        precision:
            The classifier's precision.

        Raises
        ------
        ModelNotFittedError
            If the classifier has not been fitted yet.
        """
        if not self.is_fitted:
            raise ModelNotFittedError

        validation_or_test_set = _extract_table(validation_or_test_set)

        return ClassificationMetrics.precision(
            self.predict(validation_or_test_set),
            validation_or_test_set.get_column(self.get_target_name()),
            positive_class,
        )

    def recall(self, validation_or_test_set: Table | TabularDataset, positive_class: Any) -> float:
        """
        Compute the classifier's recall on the given data.

        The recall is the proportion of actual positives that were predicted correctly. The **higher** the recall, the
        better the classifier. Results range from 0.0 to 1.0.

        **Note:** The model must be fitted.

        Parameters
        ----------
        validation_or_test_set:
            The validation or test set.
        positive_class:
            The class to be considered positive. All other classes are considered negative.

        Returns
        -------
        recall:
            The classifier's recall.

        Raises
        ------
        ModelNotFittedError
            If the classifier has not been fitted yet.
        """
        if not self.is_fitted:
            raise ModelNotFittedError

        validation_or_test_set = _extract_table(validation_or_test_set)

        return ClassificationMetrics.recall(
            self.predict(validation_or_test_set),
            validation_or_test_set.get_column(self.get_target_name()),
            positive_class,
        )


def _extract_table(table_or_dataset: Table | TabularDataset) -> Table:
    """Extract the table from the given table or dataset."""
    if isinstance(table_or_dataset, TabularDataset):
        return table_or_dataset.to_table()
    else:
        return table_or_dataset



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

from safeds._config import _get_device, _init_default_device
from safeds._utils import _structural_hash
from safeds.data.tabular.containers import Column, Table

from ._dataset import Dataset

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from torch import Tensor
    from torch.utils.data import DataLoader
    from torch.utils.data import Dataset as TorchDataset


class TabularDataset(Dataset[Table, Column]):
    """
    A dataset containing tabular data. It can be used to train machine learning models.

    Columns in a tabular dataset are divided into three categories:

    - The target column is the column that a model should predict.
    - Feature columns are columns that a model should use to make predictions.
    - Extra columns are columns that are neither feature nor target. They can be used to provide additional context,
      like an ID column.

    Feature columns are implicitly defined as all columns except the target and extra columns. If no extra columns
    are specified, all columns except the target column are used as features.

    Parameters
    ----------
    data:
        The data.
    target_name:
        The name of the target column.
    extra_names:
        Names of the columns that are neither features nor target. If None, no extra columns are used, i.e. all but
        the target column are used as features.

    Raises
    ------
    ColumnNotFoundError
        If a column name is not found in the data.
    ValueError
        If the target column is also an extra column.
    ValueError
        If no feature columns remains.

    Examples
    --------
    >>> from safeds.data.tabular.containers import Table
    >>> table = Table(
    ...     {
    ...         "id": [1, 2, 3],
    ...         "feature": [4, 5, 6],
    ...         "target": [1, 2, 3],
    ...     },
    ... )
    >>> dataset = table.to_tabular_dataset(target_name="target", extra_names=["id"])
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(
        self,
        data: Table | Mapping[str, Sequence[Any]],
        target_name: str,
        *,
        extra_names: list[str] | None = None,
    ):
        from safeds.data.tabular.containers import Table

        # Preprocess inputs
        if not isinstance(data, Table):
            data = Table(data)
        if extra_names is None:
            extra_names = []

        # Derive feature names (build the set once, since comprehensions evaluate their condition every iteration)
        non_feature_names = {target_name, *extra_names}
        feature_names = [name for name in data.column_names if name not in non_feature_names]

        # Validate inputs
        if target_name in extra_names:
            raise ValueError(f"Column '{target_name}' cannot be both target and extra.")
        if len(feature_names) == 0:
            raise ValueError("At least one feature column must remain.")

        # Set attributes
        self._table: Table = data
        self._features: Table = data.remove_columns_except(feature_names)
        self._target: Column = data.get_column(target_name)
        self._extras: Table = data.remove_columns_except(extra_names)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TabularDataset):
            return NotImplemented
        if self is other:
            return True
        return self.target == other.target and self.features == other.features and self._extras == other._extras

    def __hash__(self) -> int:
        return _structural_hash(self.target, self.features, self._extras)

    def __repr__(self) -> str:
        return self._table.__repr__()

    def __sizeof__(self) -> int:
        return sys.getsizeof(self._target) + sys.getsizeof(self._features) + sys.getsizeof(self._extras)

    def __str__(self) -> str:
        return self._table.__str__()

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def features(self) -> Table:
        """The feature columns of the tabular dataset."""
        return self._features

    @property
    def target(self) -> Column:
        """The target column of the tabular dataset."""
        return self._target

    @property
    def extras(self) -> Table:
        """
        Additional columns of the tabular dataset that are neither features nor target.

        These can be used to store additional information about instances, such as IDs.
        """
        return self._extras

    # ------------------------------------------------------------------------------------------------------------------
    # Conversion
    # ------------------------------------------------------------------------------------------------------------------

    def to_table(self) -> Table:
        """
        Return a table containing all columns of the tabular dataset.

        Returns
        -------
        table:
            A table containing all columns of the tabular dataset.
        """
        return self._table

    # ------------------------------------------------------------------------------------------------------------------
    # IPython integration
    # ------------------------------------------------------------------------------------------------------------------

    def _repr_html_(self) -> str:
        """
        Return a compact HTML representation of the tabular dataset for IPython.

        Returns
        -------
        html:
            The generated HTML.
        """
        return self._table._repr_html_()

    # TODO
    def _into_dataloader_with_classes(self, batch_size: int, num_of_classes: int) -> DataLoader:
        """
        Return a Dataloader for the data stored in this table, used for training neural networks.

        The original table is not modified.

        Parameters
        ----------
        batch_size:
            The size of data batches that should be loaded at one time.

        Returns
        -------
        result:
            The DataLoader.

        """
        import torch
        from torch.utils.data import DataLoader

        _init_default_device()

        if num_of_classes <= 2:
            return DataLoader(
                dataset=_create_dataset(
                    self.features._data_frame.to_torch().to(_get_device()),
                    self.target._series.to_torch().to(_get_device()).unsqueeze(dim=-1),
                ),
                batch_size=batch_size,
                shuffle=True,
                generator=torch.Generator(device=_get_device()),
            )
        else:
            return DataLoader(
                dataset=_create_dataset(
                    self.features._data_frame.to_torch().to(_get_device()),
                    torch.nn.functional.one_hot(
                        self.target._series.to_torch().to(_get_device()),
                        num_classes=num_of_classes,
                    ),
                ),
                batch_size=batch_size,
                shuffle=True,
                generator=torch.Generator(device=_get_device()),
            )


# TODO
def _create_dataset(features: Tensor, target: Tensor) -> TorchDataset:
    import torch
    from torch.utils.data import Dataset as TorchDataset

    _init_default_device()

    class _CustomDataset(TorchDataset):
        def __init__(self, features: Tensor, target: Tensor):
            self.X = features.to(torch.float32)
            self.Y = target.to(torch.float32)
            self.len = self.X.size(dim=0)

        def __getitem__(self, item: int) -> tuple[torch.Tensor, torch.Tensor]:
            return self.X[item], self.Y[item]

        def __len__(self) -> int:
            return self.len

    return _CustomDataset(features, target)



from __future__ import annotations

from pathlib import Path

from safeds.exceptions import FileExtensionError


def _normalize_and_check_file_path(
    path: str | Path,
    canonical_file_extension: str,
    valid_file_extensions: list[str],
    *,
    check_if_file_exists: bool = False,
) -> Path:
    """
    Check if the provided path is a valid file path and normalize it.

    Parameters
    ----------
    path:
        Path to check and normalize.
    canonical_file_extension:
        If the path has no extension, this extension will be added. Should include the leading dot.
    valid_file_extensions:
        If the path has an extension, it must be in this set. Should include the leading dots.
    check_if_file_exists:
        Whether to also check if the path points to an existing file.

    Returns
    -------
    normalized_path:
        The normalized path.

    Raises
    ------
    ValueError
        If the path has an extension that is not in the `valid_file_extensions` list.
    FileNotFoundError
        If `check_if_file_exists` is True and the file does not exist.
    """
    path = Path(path)

    # Normalize and check file extension
    if not path.suffix:
        path = path.with_suffix(canonical_file_extension)
    elif path.suffix not in valid_file_extensions:
        message = _build_file_extension_error_message(path.suffix, valid_file_extensions)
        raise FileExtensionError(message)

    # Check if file exists
    if check_if_file_exists and not path.is_file():
        raise FileNotFoundError(f"File not found: {path}")

    return path


def _build_file_extension_error_message(actual_file_extension: str, valid_file_extensions: list[str]) -> str:
    return f"Expected path with extension in {valid_file_extensions} but got {actual_file_extension}."



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _structural_hash
from safeds._validation import _check_bounds, _check_columns_exist, _ClosedBound
from safeds.data.tabular.containers import Table
from safeds.exceptions import TransformerNotFittedError

from ._table_transformer import TableTransformer

if TYPE_CHECKING:
    from sklearn.impute import KNNImputer as sk_KNNImputer


class KNearestNeighborsImputer(TableTransformer):
    """
    The KNearestNeighborsImputer replaces missing values in given Columns with the mean value of the K-nearest neighbors.

    Parameters
    ----------
    neighbor_count:
        The number of neighbors to consider when imputing missing values.
    column_names:
        The list of columns used to impute missing values. If 'None', all columns are used.
    value_to_replace:
        The placeholder for the missing values. All occurrences of`missing_values` will be imputed.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(
        self,
        neighbor_count: int,
        *,
        column_names: str | list[str] | None = None,
        value_to_replace: float | str | None = None,
    ) -> None:
        super().__init__(column_names)

        _check_bounds(name="neighbor_count", actual=neighbor_count, lower_bound=_ClosedBound(1))

        # parameter
        self._neighbor_count: int = neighbor_count
        self._value_to_replace: float | str | None = value_to_replace

        # attributes
        self._wrapped_transformer: sk_KNNImputer | None = None

    def __hash__(self) -> int:
        return _structural_hash(
            super().__hash__(),
            self._neighbor_count,
            self._value_to_replace,
            # Leave out the internal state for faster hashing
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def is_fitted(self) -> bool:
        """Whether the transformer is fitted."""
        return self._wrapped_transformer is not None

    @property
    def neighbor_count(self) -> int:
        """The number of neighbors to consider when imputing missing values."""
        return self._neighbor_count

    @property
    def value_to_replace(self) -> float | str | None:
        """The value to replace."""
        return self._value_to_replace

    # ------------------------------------------------------------------------------------------------------------------
    # Learning and transformation
    # ------------------------------------------------------------------------------------------------------------------

    def fit(self, table: Table) -> KNearestNeighborsImputer:
        """
        Learn a transformation for a set of columns in a table.

        **Note:** This transformer is not modified.

        Parameters
        ----------
        table:
            The table used to fit the transformer.

        Returns
        -------
        fitted_transformer:
            The fitted transformer.

        Raises
        ------
        ColumnNotFoundError
            If one of the columns, that should be fitted is not in the table.
        """
        from sklearn.impute import KNNImputer as sk_KNNImputer

        if table.row_count == 0:
            raise ValueError("The KNearestNeighborsImputer cannot be fitted because the table contains 0 rows.")

        if self._column_names is None:
            column_names = table.column_names
        else:
            column_names = self._column_names
            _check_columns_exist(table, column_names)

        value_to_replace = self._value_to_replace

        if self._value_to_replace is None:
            from numpy import nan

            value_to_replace = nan

        wrapped_transformer = sk_KNNImputer(n_neighbors=self._neighbor_count, missing_values=value_to_replace)
        wrapped_transformer.set_output(transform="polars")
        wrapped_transformer.fit(
            table.remove_columns_except(column_names)._data_frame,
        )

        result = KNearestNeighborsImputer(self._neighbor_count, column_names=column_names)
        result._wrapped_transformer = wrapped_transformer

        return result

    def transform(self, table: Table) -> Table:
        """
        Apply the learned transformation to a table.

        **Note:** The given table is not modified.

        Parameters
        ----------
        table:
            The table to wich the learned transformation is applied.

        Returns
        -------
        transformed_table:
            The transformed table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer is not fitted.
        ColumnNotFoundError
            If one of the columns, that should be transformed is not in the table.
        """
        if self._column_names is None or self._wrapped_transformer is None:
            raise TransformerNotFittedError

        _check_columns_exist(table, self._column_names)

        new_data = self._wrapped_transformer.transform(
            table.remove_columns_except(self._column_names)._data_frame,
        )

        return Table._from_polars_lazy_frame(
            table._lazy_frame.with_columns(new_data),
        )



from __future__ import annotations

from abc import ABC, abstractmethod

from safeds._utils import _structural_hash
from safeds._validation import _check_bounds, _ClosedBound


class _DecisionTreeBase(ABC):
    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def __init__(
        self,
        max_depth: int | None,
        min_sample_count_in_leaves: int,
    ) -> None:
        # Validation
        _check_bounds("max_depth", max_depth, lower_bound=_ClosedBound(1))
        _check_bounds(
            "min_sample_count_in_leaves",
            min_sample_count_in_leaves,
            lower_bound=_ClosedBound(1),
        )

        # Hyperparameters
        self._max_depth: int | None = max_depth
        self._min_sample_count_in_leaves: int = min_sample_count_in_leaves

    def __hash__(self) -> int:
        return _structural_hash(
            self._max_depth,
            self._min_sample_count_in_leaves,
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def max_depth(self) -> int | None:
        """The maximum depth of the tree."""
        return self._max_depth

    @property
    def min_sample_count_in_leaves(self) -> int:
        """The minimum number of samples that must remain in the leaves of the tree."""
        return self._min_sample_count_in_leaves



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._config import _init_default_device
from safeds.data.image.containers import ImageList
from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList
from safeds.data.labeled.containers import ImageDataset

from ._input_converter_image import _InputConversionImage

if TYPE_CHECKING:
    from torch import Tensor


class InputConversionImageToImage(_InputConversionImage):
    def _data_conversion_output(
        self,
        input_data: ImageList,
        output_data: Tensor,
    ) -> ImageDataset[ImageList]:
        import torch

        _init_default_device()

        if not isinstance(input_data, _SingleSizeImageList):
            raise ValueError("The given input ImageList contains images of different sizes.")  # noqa: TRY004

        return ImageDataset[ImageList](
            input_data,
            _SingleSizeImageList._create_from_tensor(
                (output_data * 255).to(torch.uint8),
                list(range(output_data.size(dim=0))),
            ),
        )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import polars as pl


def _get_polars_config() -> pl.Config:
    import polars as pl

    return pl.Config(
        float_precision=5,
        tbl_cell_numeric_alignment="RIGHT",
        tbl_formatting="ASCII_FULL_CONDENSED",
        tbl_hide_dataframe_shape=True,
    )



from __future__ import annotations

import warnings
from typing import Any

from safeds._utils import _structural_hash
from safeds._validation import _check_columns_exist
from safeds._validation._check_columns_are_numeric import _check_columns_are_numeric
from safeds.data.tabular.containers import Table
from safeds.exceptions import (
    TransformerNotFittedError,
)

from ._invertible_table_transformer import InvertibleTableTransformer


class OneHotEncoder(InvertibleTableTransformer):
    """
    A way to deal with categorical features that is particularly useful for unordered (i.e. nominal) data.

    It replaces a column with a set of columns, each representing a unique value in the original column. The value of
    each new column is 1 if the original column had that value, and 0 otherwise. Take the following table as an
    example:

    | col1 |
    |------|
    | "a"  |
    | "b"  |
    | "c"  |
    | "a"  |

    The one-hot encoding of this table is:

    | col1__a | col1__b | col1__c |
    |---------|---------|---------|
    | 1       | 0       | 0       |
    | 0       | 1       | 0       |
    | 0       | 0       | 1       |
    | 1       | 0       | 0       |

    The name "one-hot" comes from the fact that each row has exactly one 1 in it, and the rest of the values are 0s.
    One-hot encoding is closely related to dummy variable / indicator variables, which are used in statistics.

    Parameters
    ----------
    column_names:
        The list of columns used to fit the transformer. If `None`, all non-numeric columns are used.
    separator:
        The separator used to separate the original column name from the value in the new column names.

    Examples
    --------
    >>> from safeds.data.tabular.containers import Table
    >>> from safeds.data.tabular.transformation import OneHotEncoder
    >>> table = Table({"col1": ["a", "b", "c", "a"]})
    >>> transformer = OneHotEncoder()
    >>> transformer.fit_and_transform(table)[1]
    +---------+---------+---------+
    | col1__a | col1__b | col1__c |
    |     --- |     --- |     --- |
    |      u8 |      u8 |      u8 |
    +=============================+
    |       1 |       0 |       0 |
    |       0 |       1 |       0 |
    |       0 |       0 |       1 |
    |       1 |       0 |       0 |
    +---------+---------+---------+
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(
        self,
        *,
        column_names: str | list[str] | None = None,
        separator: str = "__",
    ) -> None:
        super().__init__(column_names)

        # Parameters
        self._separator = separator

        # Internal state
        self._new_column_names: list[str] | None = None
        self._mapping: dict[str, list[tuple[str, Any]]] | None = None  # Column name -> (new column name, value)[]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, OneHotEncoder):
            return NotImplemented
        elif self is other:
            return True

        return self._separator == other._separator and self._mapping == other._mapping

    def __hash__(self) -> int:
        return _structural_hash(
            super().__hash__(),
            self._separator,
            # TODO: Leave out the internal state for faster hashing
            self._mapping,
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def is_fitted(self) -> bool:
        """Whether the transformer is fitted."""
        return self._mapping is not None

    @property
    def separator(self) -> str:
        """The separator used to separate the original column name from the value in the new column names."""
        return self._separator

    # ------------------------------------------------------------------------------------------------------------------
    # Learning and transformation
    # ------------------------------------------------------------------------------------------------------------------

    def fit(self, table: Table) -> OneHotEncoder:
        """
        Learn a transformation for a set of columns in a table.

        This transformer is not modified.

        Parameters
        ----------
        table:
            The table used to fit the transformer.

        Returns
        -------
        fitted_transformer:
            The fitted transformer.

        Raises
        ------
        ColumnNotFoundError
            If column_names contain a column name that is missing in the table.
        ValueError
            If the table contains 0 rows.
        """
        if self._column_names is None:
            column_names = [name for name in table.column_names if not table.get_column_type(name).is_numeric]
        else:
            column_names = self._column_names
            _check_columns_exist(table, column_names)
            _warn_if_columns_are_numeric(table, column_names)

        if table.row_count == 0:
            raise ValueError("The OneHotEncoder cannot be fitted because the table contains 0 rows")

        # Learn the transformation
        new_column_names: list[str] = []
        mapping: dict[str, list[tuple[str, Any]]] = {}

        known_names = set(table.column_names)

        for name in column_names:
            mapping[name] = []
            for value in table.get_column(name).get_distinct_values():
                base_name = f"{name}{self._separator}{value}"
                new_name = base_name

                # Ensure that the new column name is unique
                counter = 2
                while new_name in known_names:
                    new_name = f"{base_name}#{counter}"
                    counter += 1

                known_names.add(new_name)
                new_column_names.append(new_name)
                mapping[name].append((new_name, value))

        # Create a copy with the learned transformation
        result = OneHotEncoder(column_names=column_names, separator=self._separator)
        result._new_column_names = new_column_names
        result._mapping = mapping

        return result

    def transform(self, table: Table) -> Table:
        """
        Apply the learned transformation to a table.

        The table is not modified.

        Parameters
        ----------
        table:
            The table to which the learned transformation is applied.

        Returns
        -------
        transformed_table:
            The transformed table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        ColumnNotFoundError
            If the input table does not contain all columns used to fit the transformer.
        """
        import polars as pl

        # Used in favor of is_fitted, so the type checker is happy
        if self._column_names is None or self._mapping is None:
            raise TransformerNotFittedError

        _check_columns_exist(table, self._column_names)

        expressions = [
            # UInt8 can be used without conversion in scikit-learn
            pl.col(column_name).eq_missing(value).alias(new_name).cast(pl.UInt8)
            for column_name in self._column_names
            for new_name, value in self._mapping[column_name]
        ]

        return Table._from_polars_lazy_frame(
            table._lazy_frame.with_columns(expressions).drop(self._column_names),
        )

    def inverse_transform(self, transformed_table: Table) -> Table:
        """
        Undo the learned transformation.

        The table is not modified.

        Parameters
        ----------
        transformed_table:
            The table to be transformed back to the original version.

        Returns
        -------
        table:
            The original table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        ColumnNotFoundError
            If the input table does not contain all columns used to fit the transformer.
        NonNumericColumnError
            If the transformed columns of the input table contain non-numerical data.
        """
        import polars as pl

        # Used in favor of is_fitted, so the type checker is happy
        if self._column_names is None or self._new_column_names is None or self._mapping is None:
            raise TransformerNotFittedError

        _check_columns_exist(transformed_table, self._new_column_names)
        _check_columns_are_numeric(
            transformed_table,
            self._new_column_names,
            operation="inverse-transform with a OneHotEncoder",
        )

        expressions = [
            pl.coalesce(
                [
                    # The pl.lit is needed, so strings are not interpreted as column names
                    pl.when(pl.col(new_column_name) == 1).then(pl.lit(value))
                    for new_column_name, value in self._mapping[column_name]
                ],
            ).alias(column_name)
            for column_name in self._mapping
        ]

        return Table._from_polars_lazy_frame(
            transformed_table._lazy_frame.with_columns(expressions).drop(self._new_column_names),
        )

    # TODO: remove / replace with consistent introspection methods across all transformers
    def _get_names_of_added_columns(self) -> list[str]:
        if self._new_column_names is None:
            raise TransformerNotFittedError
        return list(self._new_column_names)  # defensive copy


def _warn_if_columns_are_numeric(table: Table, column_names: list[str]) -> None:
    numeric_columns = table.remove_columns_except(column_names).remove_non_numeric_columns().column_names
    if numeric_columns:
        warnings.warn(
            f"The columns {numeric_columns} contain numerical data. "
            "The OneHotEncoder is designed to encode non-numerical values into numerical values",
            UserWarning,
            stacklevel=2,
        )



"""Tools to work with hyperparameters of ML models."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._choice import Choice

apipkg.initpkg(
    __name__,
    {
        "Choice": "._choice:Choice",
    },
)

__all__ = ["Choice"]



"""Classes to evaluate the performance of machine learning models."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._classification_metrics import ClassificationMetrics
    from ._regression_metrics import RegressionMetrics

apipkg.initpkg(
    __name__,
    {
        "ClassificationMetrics": "._classification_metrics:ClassificationMetrics",
        "RegressionMetrics": "._regression_metrics:RegressionMetrics",
    },
)

__all__ = [
    "ClassificationMetrics",
    "RegressionMetrics",
]



"""Classes for classification tasks."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._ada_boost_classifier import AdaBoostClassifier
    from ._baseline_classifier import BaselineClassifier
    from ._classifier import Classifier
    from ._decision_tree_classifier import DecisionTreeClassifier
    from ._gradient_boosting_classifier import GradientBoostingClassifier
    from ._k_nearest_neighbors_classifier import KNearestNeighborsClassifier
    from ._logistic_classifier import LogisticClassifier
    from ._random_forest_classifier import RandomForestClassifier
    from ._support_vector_classifier import SupportVectorClassifier

apipkg.initpkg(
    __name__,
    {
        "AdaBoostClassifier": "._ada_boost_classifier:AdaBoostClassifier",
        "BaselineClassifier": "._baseline_classifier:BaselineClassifier",
        "Classifier": "._classifier:Classifier",
        "DecisionTreeClassifier": "._decision_tree_classifier:DecisionTreeClassifier",
        "GradientBoostingClassifier": "._gradient_boosting_classifier:GradientBoostingClassifier",
        "KNearestNeighborsClassifier": "._k_nearest_neighbors_classifier:KNearestNeighborsClassifier",
        "LogisticClassifier": "._logistic_classifier:LogisticClassifier",
        "RandomForestClassifier": "._random_forest_classifier:RandomForestClassifier",
        "SupportVectorClassifier": "._support_vector_classifier:SupportVectorClassifier",
    },
)

__all__ = [
    "AdaBoostClassifier",
    "BaselineClassifier",
    "Classifier",
    "DecisionTreeClassifier",
    "GradientBoostingClassifier",
    "KNearestNeighborsClassifier",
    "LogisticClassifier",
    "RandomForestClassifier",
    "SupportVectorClassifier",
]



"""Custom exceptions that can be raised by Safe-DS."""

from ._data import (
    ColumnLengthMismatchError,
    ColumnSizeError,
    DuplicateColumnError,
    DuplicateIndexError,
    IllegalFormatError,
    IndexOutOfBoundsError,
    MissingValuesColumnError,
    NonNumericColumnError,
    OutputLengthMismatchError,
    TransformerNotFittedError,
    TransformerNotInvertibleError,
    ValueNotPresentWhenFittedError,
)
from ._ml import (
    DatasetMissesDataError,
    DatasetMissesFeaturesError,
    EmptyChoiceError,
    FeatureDataMismatchError,
    FittingWithChoiceError,
    FittingWithoutChoiceError,
    InputSizeError,
    InvalidFitDataError,
    InvalidModelStructureError,
    LearningError,
    ModelNotFittedError,
    PlainTableError,
    PredictionError,
    TargetDataMismatchError,
)


class SafeDsError(Exception):
    """Base class for all exceptions defined by Safe-DS."""


class ColumnNotFoundError(SafeDsError):
    """Exception raised when trying to access an invalid column name."""


class ColumnTypeError(SafeDsError):
    """Exception raised when a column has the wrong type."""


class FileExtensionError(SafeDsError):
    """Exception raised when a path has the wrong file extension."""


class OutOfBoundsError(SafeDsError):
    """Exception raised when a value is outside its expected range."""


__all__ = [
    "SafeDsError",
    "ColumnNotFoundError",
    "ColumnTypeError",
    "FileExtensionError",
    "OutOfBoundsError",
    # TODO
    # Data exceptions
    "ColumnLengthMismatchError",
    "ColumnSizeError",
    "DuplicateColumnError",
    "DuplicateIndexError",
    "IllegalFormatError",
    "IndexOutOfBoundsError",
    "MissingValuesColumnError",
    "NonNumericColumnError",
    "OutputLengthMismatchError",
    "TransformerNotFittedError",
    "TransformerNotInvertibleError",
    "ValueNotPresentWhenFittedError",
    # ML exceptions
    "DatasetMissesDataError",
    "DatasetMissesFeaturesError",
    "EmptyChoiceError",
    "FeatureDataMismatchError",
    "FittingWithChoiceError",
    "FittingWithoutChoiceError",
    "InvalidFitDataError",
    "InputSizeError",
    "InvalidModelStructureError",
    "LearningError",
    "ModelNotFittedError",
    "PlainTableError",
    "PredictionError",
    "TargetDataMismatchError",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _structural_hash
from safeds.ml.classical._bases import _KNearestNeighborsBase

from ._classifier import Classifier

if TYPE_CHECKING:
    from sklearn.base import ClassifierMixin

    from safeds.data.labeled.containers import TabularDataset


class KNearestNeighborsClassifier(Classifier, _KNearestNeighborsBase):
    """
    K-nearest-neighbors classification.

    Parameters
    ----------
    neighbor_count:
        The number of neighbors to use for interpolation. Has to be greater than 0 (validated in the constructor) and
        less than or equal to the sample size (validated when calling `fit`).

    Raises
    ------
    OutOfBoundsError
        If `neighbor_count` is less than 1.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(
        self,
        neighbor_count: int,
    ) -> None:
        # Initialize superclasses
        Classifier.__init__(self)
        _KNearestNeighborsBase.__init__(
            self,
            neighbor_count=neighbor_count,
        )

    def __hash__(self) -> int:
        return _structural_hash(
            Classifier.__hash__(self),
            _KNearestNeighborsBase.__hash__(self),
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Template methods
    # ------------------------------------------------------------------------------------------------------------------

    def _check_additional_fit_preconditions(self, training_set: TabularDataset) -> None:
        if self._neighbor_count > training_set._table.row_count:
            raise ValueError(
                (
                    f"The parameter 'neighbor_count' ({self._neighbor_count}) has to be less than or equal to"
                    f" the sample size ({training_set._table.row_count})."
                ),
            )

    def _clone(self) -> KNearestNeighborsClassifier:
        return KNearestNeighborsClassifier(
            neighbor_count=self._neighbor_count,
        )

    def _get_sklearn_model(self) -> ClassifierMixin:
        from sklearn.neighbors import KNeighborsClassifier as SklearnKNeighborsClassifier

        return SklearnKNeighborsClassifier(
            n_neighbors=self._neighbor_count,
            n_jobs=-1,
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from warnings import warn

from safeds._utils import _structural_hash
from safeds.exceptions import TransformerNotFittedError, TransformerNotInvertibleError

from ._invertible_table_transformer import InvertibleTableTransformer

if TYPE_CHECKING:
    from safeds.data.tabular.containers import Table

    from ._table_transformer import TableTransformer


class SequentialTableTransformer(InvertibleTableTransformer):
    """
    The SequentialTableTransformer transforms a table using multiple transformers in sequence.

    Parameters
    ----------
    transformers:
        The list of transformers used to transform the table. Used in the order as they are supplied in the list.
    """

    def __init__(
        self,
        transformers: list[TableTransformer],
    ) -> None:
        super().__init__(None)

        # Check if transformers actually contains any transformers.
        if transformers is None or len(transformers) == 0:
            warn(
                "transformers should contain at least 1 transformer",
                UserWarning,
                stacklevel=2,
            )

        # Parameters
        self._transformers: list[TableTransformer] = transformers

        # Internal State
        self._is_fitted: bool = False

    def __hash__(self) -> int:
        return _structural_hash(
            super().__hash__(),
            self._transformers,
            self._is_fitted,
        )

    @property
    def is_fitted(self) -> bool:
        """Whether the transformer is fitted."""
        return self._is_fitted

    def fit(self, table: Table) -> SequentialTableTransformer:
        """
        Fits all the transformers in order.

        Parameters
        ----------
        table:
            The table used to fit the transformers.

        Returns
        -------
        fitted_transformer:
            The fitted transformer.

        Raises
        ------
        ValueError:
            Raises a ValueError if the table has no rows.
        """
        if table.row_count == 0:
            raise ValueError("The SequentialTableTransformer cannot be fitted because the table contains 0 rows.")

        current_table: Table = table
        fitted_transformers: list[TableTransformer] = []

        for transformer in self._transformers:
            fitted_transformer = transformer.fit(current_table)
            fitted_transformers.append(fitted_transformer)
            current_table = fitted_transformer.transform(current_table)

        result: SequentialTableTransformer = SequentialTableTransformer(
            transformers=fitted_transformers,
        )

        result._is_fitted = True
        return result

    def transform(self, table: Table) -> Table:
        """
        Transform the table using all the transformers sequentially.

        Might change the order and type of columns base on the transformers used.

        Parameters
        ----------
        table:
            The table to be transformed.

        Returns
        -------
        transformed_table:
            The transformed table.

        Raises
        ------
        TransformerNotFittedError:
            Raises a TransformerNotFittedError if the transformer isn't fitted.
        """
        if not self._is_fitted:
            raise TransformerNotFittedError

        current_table: Table = table
        for transformer in self._transformers:
            current_table = transformer.transform(current_table)

        return current_table

    def inverse_transform(self, transformed_table: Table) -> Table:
        """
        Inversely transforms the table using all the transformers sequentially in inverse order.

        Might change the order and type of columns base on the transformers used.

        Parameters
        ----------
        transformed_table:
            The table to be transformed back.

        Returns
        -------
        original_table:
            The original table.

        Raises
        ------
        TransformerNotFittedError:
            Raises a TransformerNotFittedError if the transformer isn't fitted.
        TransformerNotInvertibleError:
            Raises a TransformerNotInvertibleError if one of the transformers isn't invertible.
        """
        if not self._is_fitted:
            raise TransformerNotFittedError

        # sequentially inverse transform the table with all transformers, working from the back of the list forwards.
        current_table: Table = transformed_table
        for transformer in reversed(self._transformers):
            # check if transformer is invertible
            if not (isinstance(transformer, InvertibleTableTransformer)):
                raise TransformerNotInvertibleError(str(type(transformer)))
            current_table = transformer.inverse_transform(current_table)

        return current_table



"""Base classes for classical machine learning models."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._ada_boost_base import _AdaBoostBase
    from ._decision_tree_base import _DecisionTreeBase
    from ._gradient_boosting_base import _GradientBoostingBase
    from ._k_nearest_neighbors_base import _KNearestNeighborsBase
    from ._random_forest_base import _RandomForestBase
    from ._support_vector_machine_base import _SupportVectorMachineBase

apipkg.initpkg(
    __name__,
    {
        "_AdaBoostBase": "._ada_boost_base:_AdaBoostBase",
        "_DecisionTreeBase": "._decision_tree_base:_DecisionTreeBase",
        "_GradientBoostingBase": "._gradient_boosting_base:_GradientBoostingBase",
        "_KNearestNeighborsBase": "._k_nearest_neighbors_base:_KNearestNeighborsBase",
        "_RandomForestBase": "._random_forest_base:_RandomForestBase",
        "_SupportVectorMachineBase": "._support_vector_machine_base:_SupportVectorMachineBase",
    },
)

__all__ = [
    "_AdaBoostBase",
    "_DecisionTreeBase",
    "_GradientBoostingBase",
    "_KNearestNeighborsBase",
    "_RandomForestBase",
    "_SupportVectorMachineBase",
]



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING

from safeds.data.labeled.containers import TabularDataset
from safeds.exceptions import ColumnLengthMismatchError, ModelNotFittedError
from safeds.ml.classical import SupervisedModel
from safeds.ml.metrics import RegressionMetrics

if TYPE_CHECKING:
    from safeds.data.tabular.containers import Column, Table


class Regressor(SupervisedModel, ABC):
    """A model for regression tasks."""

    # ------------------------------------------------------------------------------------------------------------------
    # Metrics
    # ------------------------------------------------------------------------------------------------------------------

    def summarize_metrics(self, validation_or_test_set: Table | TabularDataset) -> Table:
        """
        Summarize the regressor's metrics on the given data.

        **Note:** The model must be fitted.

        Parameters
        ----------
        validation_or_test_set:
            The validation or test set.

        Returns
        -------
        metrics:
            A table containing the regressor's metrics.

        Raises
        ------
        ModelNotFittedError
            If the classifier has not been fitted yet.
        """
        if not self.is_fitted:
            raise ModelNotFittedError

        validation_or_test_set = _extract_table(validation_or_test_set)

        return RegressionMetrics.summarize(
            self.predict(validation_or_test_set),
            validation_or_test_set.get_column(self.get_target_name()),
        )

    def coefficient_of_determination(self, validation_or_test_set: Table | TabularDataset) -> float:
        """
        Compute the coefficient of determination (R²) of the regressor on the given data.

        The coefficient of determination compares the regressor's predictions to another model that always predicts the
        mean of the target values. It is a measure of how well the regressor explains the variance in the target values.

        The **higher** the coefficient of determination, the better the regressor. Results range from negative infinity
        to 1.0. You can interpret the coefficient of determination as follows:

        | R²         | Interpretation                                                                             |
        | ---------- | ------------------------------------------------------------------------------------------ |
        | 1.0        | The model perfectly predicts the target values. Did you overfit?                           |
        | (0.0, 1.0) | The model is better than predicting the mean of the target values. You should be here.     |
        | 0.0        | The model is as good as predicting the mean of the target values. Try something else.      |
        | (-∞, 0.0)  | The model is worse than predicting the mean of the target values. Something is very wrong. |

        **Notes:**

        - The model must be fitted.
        - Some other libraries call this metric `r2_score`.

        Parameters
        ----------
        validation_or_test_set:
            The validation or test set.

        Returns
        -------
        coefficient_of_determination:
            The coefficient of determination of the regressor.

        Raises
        ------
        ModelNotFittedError
            If the classifier has not been fitted yet.
        """
        if not self.is_fitted:
            raise ModelNotFittedError

        validation_or_test_set = _extract_table(validation_or_test_set)

        return RegressionMetrics.coefficient_of_determination(
            self.predict(validation_or_test_set),
            validation_or_test_set.get_column(self.get_target_name()),
        )

    def mean_absolute_error(self, validation_or_test_set: Table | TabularDataset) -> float:
        """
        Compute the mean absolute error (MAE) of the regressor on the given data.

        The mean absolute error is the average of the absolute differences between the predicted and expected target
        values. The **lower** the mean absolute error, the better the regressor. Results range from 0.0 to positive
        infinity.

        **Note:** The model must be fitted.

        Parameters
        ----------
        validation_or_test_set:
            The validation or test set.

        Returns
        -------
        mean_absolute_error:
            The mean absolute error of the regressor.

        Raises
        ------
        ModelNotFittedError
            If the classifier has not been fitted yet.
        """
        if not self.is_fitted:
            raise ModelNotFittedError

        validation_or_test_set = _extract_table(validation_or_test_set)

        return RegressionMetrics.mean_absolute_error(
            self.predict(validation_or_test_set),
            validation_or_test_set.get_column(self.get_target_name()),
        )

    def mean_directional_accuracy(self, validation_or_test_set: Table | TabularDataset) -> float:
        """
        Compute the mean directional accuracy (MDA) of the regressor on the given data.

        This metric compares two consecutive target values and checks if the predicted direction (down/unchanged/up)
        matches the expected direction. The mean directional accuracy is the proportion of correctly predicted
        directions. The **higher** the mean directional accuracy, the better the regressor. Results range from 0.0 to
        1.0.

        This metric is useful for time series data, where the order of the target values has a meaning. It is not useful
        for other types of data. Because of this, it is not included in the `summarize_metrics` method.

        **Note:** The model must be fitted.

        Parameters
        ----------
        validation_or_test_set:
            The validation or test set.

        Returns
        -------
        mean_directional_accuracy:
            The mean directional accuracy of the regressor.

        Raises
        ------
        ModelNotFittedError
            If the classifier has not been fitted yet.
        """
        if not self.is_fitted:
            raise ModelNotFittedError

        validation_or_test_set = _extract_table(validation_or_test_set)

        return RegressionMetrics.mean_directional_accuracy(
            self.predict(validation_or_test_set),
            validation_or_test_set.get_column(self.get_target_name()),
        )

    def mean_squared_error(self, validation_or_test_set: Table | TabularDataset) -> float:
        """
        Compute the mean squared error (MSE) of the regressor on the given data.

        The mean squared error is the average of the squared differences between the predicted and expected target
        values. The **lower** the mean squared error, the better the regressor. Results range from 0.0 to positive
        infinity.

        **NoteS:**

        - The model must be fitted.
        - To get the root mean squared error (RMSE), take the square root of the result.

        Parameters
        ----------
        validation_or_test_set:
            The validation or test set.

        Returns
        -------
        mean_squared_error:
            The mean squared error of the regressor.

        Raises
        ------
        ModelNotFittedError
            If the classifier has not been fitted yet.
        """
        if not self.is_fitted:
            raise ModelNotFittedError

        validation_or_test_set = _extract_table(validation_or_test_set)

        return RegressionMetrics.mean_squared_error(
            self.predict(validation_or_test_set),
            validation_or_test_set.get_column(self.get_target_name()),
        )

    def median_absolute_deviation(self, validation_or_test_set: Table | TabularDataset) -> float:
        """
        Compute the median absolute deviation (MAD) of the regressor on the given data.

        The median absolute deviation is the median of the absolute differences between the predicted and expected
        target values. The **lower** the median absolute deviation, the better the regressor. Results range from 0.0 to
        positive infinity.

        **Note:** The model must be fitted.

        Parameters
        ----------
        validation_or_test_set:
            The validation or test set.

        Returns
        -------
        median_absolute_deviation:
            The median absolute deviation of the regressor.

        Raises
        ------
        ModelNotFittedError
            If the classifier has not been fitted yet.
        """
        if not self.is_fitted:
            raise ModelNotFittedError

        validation_or_test_set = _extract_table(validation_or_test_set)

        return RegressionMetrics.median_absolute_deviation(
            self.predict(validation_or_test_set),
            validation_or_test_set.get_column(self.get_target_name()),
        )


def _check_metrics_preconditions(actual: Column, expected: Column) -> None:  # pragma: no cover
    if not actual.type.is_numeric:
        raise TypeError(f"Column 'actual' is not numerical but {actual.type}.")
    if not expected.type.is_numeric:
        raise TypeError(f"Column 'expected' is not numerical but {expected.type}.")

    if actual.row_count != expected.row_count:
        raise ColumnLengthMismatchError(
            "\n".join(
                [
                    f"{actual.name}: {actual.row_count}",
                    f"{expected.name}: {expected.row_count}",
                ],
            ),
        )


def _extract_table(table_or_dataset: Table | TabularDataset) -> Table:
    """Extract the table from the given table or dataset."""
    if isinstance(table_or_dataset, TabularDataset):
        return table_or_dataset.to_table()
    else:
        return table_or_dataset



from __future__ import annotations

import sys
from abc import ABC
from typing import TYPE_CHECKING

from safeds._utils import _structural_hash
from safeds.data.image.containers import ImageList
from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList
from safeds.data.labeled.containers import ImageDataset
from safeds.data.labeled.containers._image_dataset import _ColumnAsTensor, _TableAsTensor

from ._input_converter import InputConversion

if TYPE_CHECKING:
    from safeds.data.tabular.transformation import OneHotEncoder
    from safeds.ml.nn.typing import ModelImageSize


class _InputConversionImage(InputConversion[ImageDataset, ImageList], ABC):
    """
    The input conversion for a neural network, defines the input parameters for the neural network.

    Parameters
    ----------
    image_size:
        the size of the input images
    """

    def __init__(self, image_size: ModelImageSize) -> None:
        self._input_size = image_size
        self._output_size: ModelImageSize | int | None = None
        self._one_hot_encoder: OneHotEncoder | None = None
        self._column_name: str | None = None
        self._column_names: list[str] | None = None
        self._output_type: type | None = None

    def __hash__(self) -> int:
        return _structural_hash(
            self.__class__.__name__,
            self._input_size,
            self._output_size,
            self._one_hot_encoder,
            self._column_name,
            self._column_names,
            self._output_type,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self is other) or (
            self._input_size == other._input_size
            and self._output_size == other._output_size
            and self._one_hot_encoder == other._one_hot_encoder
            and self._column_name == other._column_name
            and self._column_names == other._column_names
            and self._output_type == other._output_type
        )

    def __sizeof__(self) -> int:
        return (
            sys.getsizeof(self._input_size)
            + sys.getsizeof(self._output_size)
            + sys.getsizeof(self._one_hot_encoder)
            + sys.getsizeof(self._column_name)
            + sys.getsizeof(self._column_names)
            + sys.getsizeof(self._output_type)
        )

    @property
    def _data_size(self) -> ModelImageSize:
        return self._input_size

    def _data_conversion_fit(
        self,
        input_data: ImageDataset,
        batch_size: int,  # noqa: ARG002
        num_of_classes: int = 1,  # noqa: ARG002
    ) -> ImageDataset:
        return input_data

    def _data_conversion_predict(self, input_data: ImageList, batch_size: int) -> _SingleSizeImageList:  # noqa: ARG002
        return input_data._as_single_size_image_list()

    def _is_fit_data_valid(self, input_data: ImageDataset) -> bool:
        if self._output_type is None:
            self._output_type = type(input_data._output)
            self._output_size = input_data.output_size
        elif not isinstance(input_data._output, self._output_type):
            return False
        if isinstance(input_data._output, _ColumnAsTensor):
            if self._column_name is None and self._one_hot_encoder is None:
                self._one_hot_encoder = input_data._output._one_hot_encoder
                self._column_name = input_data._output._column_name
            elif (
                self._column_name != input_data._output._column_name
                or self._one_hot_encoder != input_data._output._one_hot_encoder
            ):
                return False
        elif isinstance(input_data._output, _TableAsTensor):
            if self._column_names is None:
                self._column_names = input_data._output._column_names
            elif self._column_names != input_data._output._column_names:
                return False
        return input_data.input_size == self._input_size and input_data.output_size == self._output_size

    def _is_predict_data_valid(self, input_data: ImageList) -> bool:
        return isinstance(input_data, _SingleSizeImageList) and input_data.sizes[0] == self._input_size



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._config import _init_default_device
from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList
from safeds.data.labeled.containers import ImageDataset
from safeds.data.labeled.containers._image_dataset import _ColumnAsTensor
from safeds.data.tabular.containers import Column

from ._input_converter_image import _InputConversionImage

if TYPE_CHECKING:
    from torch import Tensor

    from safeds.data.image.containers import ImageList
    from safeds.data.tabular.transformation import OneHotEncoder


class InputConversionImageToColumn(_InputConversionImage):
    def _data_conversion_output(
        self,
        input_data: ImageList,
        output_data: Tensor,
    ) -> ImageDataset[Column]:
        import torch

        _init_default_device()

        if not isinstance(input_data, _SingleSizeImageList):
            raise ValueError("The given input ImageList contains images of different sizes.")  # noqa: TRY004
        if self._column_name is None:
            raise ValueError(
                "The column_name is not set. The data can only be converted if the column_name is provided as `str` in the kwargs.",
            )
        if self._one_hot_encoder is None:
            raise ValueError(
                "The one_hot_encoder is not set. The data can only be converted if the one_hot_encoder is provided as `OneHotEncoder` in the kwargs.",
            )
        one_hot_encoder: OneHotEncoder = self._one_hot_encoder
        column_name: str = self._column_name

        output = torch.zeros(len(input_data), len(one_hot_encoder._get_names_of_added_columns()))
        output[torch.arange(len(input_data)), output_data] = 1

        im_dataset: ImageDataset[Column] = ImageDataset[Column].__new__(ImageDataset)
        im_dataset._output = _ColumnAsTensor._from_tensor(output, column_name, one_hot_encoder)
        im_dataset._shuffle_tensor_indices = torch.LongTensor(list(range(len(input_data))))
        im_dataset._shuffle_after_epoch = False
        im_dataset._batch_size = 1
        im_dataset._next_batch_index = 0
        im_dataset._input_size = input_data.sizes[0]
        im_dataset._input = input_data
        return im_dataset



from __future__ import annotations

import io
from typing import TYPE_CHECKING

from safeds.data.image.containers import Image

if TYPE_CHECKING:
    import matplotlib.pyplot as plt


def _figure_to_image(figure: plt.Figure) -> Image:
    """
    Store the figure as an image and closes it.

    Parameters
    ----------
    figure:
        The figure to store.

    Returns
    -------
    image:
        The figure as an image.
    """
    import matplotlib.pyplot as plt

    buffer = io.BytesIO()
    figure.savefig(buffer, format="png")
    plt.close(figure)  # Prevents the figure from being displayed directly
    buffer.seek(0)
    return Image.from_bytes(buffer.read())



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

from safeds._utils import _structural_hash
from safeds._validation import _check_bounds, _ClosedBound
from safeds.ml.hyperparameters import Choice
from safeds.ml.nn.typing import ModelImageSize

from ._layer import Layer

if TYPE_CHECKING:
    from torch import nn


class GRULayer(Layer):
    """
    A gated recurrent unit (GRU) layer.

    Parameters
    ----------
    neuron_count:
        The number of neurons in this layer

    Raises
    ------
    OutOfBoundsError
        If input_size < 1
        If output_size < 1
    """

    def __init__(self, neuron_count: int | Choice[int]):
        if isinstance(neuron_count, Choice):
            for val in neuron_count:
                _check_bounds("neuron_count", val, lower_bound=_ClosedBound(1))
        else:
            _check_bounds("neuron_count", neuron_count, lower_bound=_ClosedBound(1))

        self._input_size: int | None = None
        self._output_size = neuron_count

    def _get_internal_layer(self, **kwargs: Any) -> nn.Module:
        assert not self._contains_choices()
        assert not isinstance(self._output_size, Choice)  # just for linter
        from ._internal_layers import _InternalGRULayer  # Slow import on global level

        if "activation_function" not in kwargs:
            raise ValueError(
                "The activation_function is not set. The internal layer can only be created when the activation_function is provided in the kwargs.",
            )
        else:
            activation_function: str = kwargs["activation_function"]

        if self._input_size is None:
            raise ValueError("The input_size is not yet set.")

        return _InternalGRULayer(self._input_size, self._output_size, activation_function)

    @property
    def input_size(self) -> int:
        """
        Get the input_size of this layer.

        Returns
        -------
        result:
            The amount of values being passed into this layer.
        """
        if self._input_size is None:
            raise ValueError("The input_size is not yet set.")
        return self._input_size

    @property
    def output_size(self) -> int | Choice[int]:
        """
        Get the output_size of this layer.

        Returns
        -------
        result:
            The number of neurons in this layer.
        """
        return self._output_size

    def _set_input_size(self, input_size: int | ModelImageSize) -> None:
        if isinstance(input_size, ModelImageSize):
            raise TypeError("The input_size of a gru layer has to be of type int.")

        self._input_size = input_size

    def _contains_choices(self) -> bool:
        return isinstance(self._output_size, Choice)

    def _get_layers_for_all_choices(self) -> list[GRULayer]:
        assert self._contains_choices()
        assert isinstance(self._output_size, Choice)  # just for linter
        layers = []
        for val in self._output_size:
            layers.append(GRULayer(neuron_count=val))
        return layers

    def __hash__(self) -> int:
        return _structural_hash(
            self._input_size,
            self._output_size,
        )  # pragma: no cover

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GRULayer):
            return NotImplemented
        return (self is other) or (self._input_size == other._input_size and self._output_size == other._output_size)

    def __sizeof__(self) -> int:
        return sys.getsizeof(self._input_size) + sys.getsizeof(self._output_size)



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _get_random_seed, _structural_hash
from safeds.ml.classical._bases import _RandomForestBase

from ._classifier import Classifier

if TYPE_CHECKING:
    from sklearn.base import ClassifierMixin


class RandomForestClassifier(Classifier, _RandomForestBase):
    """
    Random forest classification.

    Parameters
    ----------
    tree_count:
        The number of trees to be used in the random forest. Has to be greater than 0.
    max_depth:
        The maximum depth of each tree. If None, the depth is not limited. Has to be greater than 0.
    min_sample_count_in_leaves:
        The minimum number of samples that must remain in the leaves of each tree. Has to be greater than 0.

    Raises
    ------
    OutOfBoundsError
        If `tree_count` is less than 1.
    OutOfBoundsError
        If `max_depth` is less than 1.
    OutOfBoundsError
        If `min_sample_count_in_leaves` is less than 1.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(
        self,
        *,
        tree_count: int = 100,
        max_depth: int | None = None,
        min_sample_count_in_leaves: int = 1,
    ) -> None:
        # Initialize superclasses
        Classifier.__init__(self)
        _RandomForestBase.__init__(
            self,
            tree_count=tree_count,
            max_depth=max_depth,
            min_sample_count_in_leaves=min_sample_count_in_leaves,
        )

    def __hash__(self) -> int:
        return _structural_hash(
            Classifier.__hash__(self),
            _RandomForestBase.__hash__(self),
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Template methods
    # ------------------------------------------------------------------------------------------------------------------

    def _clone(self) -> RandomForestClassifier:
        return RandomForestClassifier(
            tree_count=self._tree_count,
            max_depth=self._max_depth,
            min_sample_count_in_leaves=self._min_sample_count_in_leaves,
        )

    def _get_sklearn_model(self) -> ClassifierMixin:
        from sklearn.ensemble import RandomForestClassifier as SklearnRandomForestClassifier

        return SklearnRandomForestClassifier(
            n_estimators=self._tree_count,
            max_depth=self._max_depth,
            min_samples_leaf=self._min_sample_count_in_leaves,
            random_state=_get_random_seed(),
            n_jobs=-1,
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from safeds.data.labeled.containers import TabularDataset
from safeds.data.tabular.containers import Table
from safeds.exceptions import ColumnLengthMismatchError

if TYPE_CHECKING:
    from safeds.data.tabular.containers import Column


class ClassificationMetrics(ABC):
    """A collection of classification metrics."""

    @abstractmethod
    def __init__(self) -> None: ...

    @staticmethod
    def summarize(predicted: Column | TabularDataset, expected: Column | TabularDataset, positive_class: Any) -> Table:
        """
        Summarize classification metrics on the given data.

        Parameters
        ----------
        predicted:
            The predicted target values produced by the classifier.
        expected:
            The expected target values.
        positive_class:
            The class to be considered positive. All other classes are considered negative.

        Returns
        -------
        metrics:
            A table containing the classification metrics.
        """
        expected = _extract_target(expected)
        predicted = _extract_target(predicted)
        _check_equal_length(predicted, expected)

        accuracy = ClassificationMetrics.accuracy(predicted, expected)
        precision = ClassificationMetrics.precision(predicted, expected, positive_class)
        recall = ClassificationMetrics.recall(predicted, expected, positive_class)
        f1_score = ClassificationMetrics.f1_score(predicted, expected, positive_class)

        return Table(
            {
                "metric": ["accuracy", "precision", "recall", "f1_score"],
                "value": [accuracy, precision, recall, f1_score],
            },
        )

    @staticmethod
    def accuracy(predicted: Column | TabularDataset, expected: Column | TabularDataset) -> float:
        """
        Compute the accuracy on the given data.

        The accuracy is the proportion of predicted target values that were correct. The **higher** the accuracy, the
        better. Results range from 0.0 to 1.0.

        Parameters
        ----------
        predicted:
            The predicted target values produced by the classifier.
        expected:
            The expected target values.

        Returns
        -------
        accuracy:
            The calculated accuracy.
        """
        expected = _extract_target(expected)
        predicted = _extract_target(predicted)
        _check_equal_length(predicted, expected)

        from polars.exceptions import ComputeError

        if expected.row_count == 0:
            return 1.0  # Everything was predicted correctly (since there is nothing to predict)

        try:
            return expected._series.eq_missing(predicted._series).mean()
        except ComputeError:
            return 0.0  # Types are not compatible, so no prediction can be correct

    @staticmethod
    def f1_score(predicted: Column | TabularDataset, expected: Column | TabularDataset, positive_class: Any) -> float:
        """
        Compute the F₁ score on the given data.

        The F₁ score is the harmonic mean of precision and recall. The **higher** the F₁ score, the better the
        classifier. Results range from 0.0 to 1.0.

        Parameters
        ----------
        predicted:
            The predicted target values produced by the classifier.
        expected:
            The expected target values.
        positive_class:
            The class to be considered positive. All other classes are considered negative.

        Returns
        -------
        f1_score:
            The calculated F₁ score.
        """
        predicted = _extract_target(predicted)
        expected = _extract_target(expected)
        _check_equal_length(predicted, expected)

        true_positives = (expected._series.eq(positive_class) & predicted._series.eq(positive_class)).sum()
        false_positives = (expected._series.ne(positive_class) & predicted._series.eq(positive_class)).sum()
        false_negatives = (expected._series.eq(positive_class) & predicted._series.ne(positive_class)).sum()

        if true_positives + false_positives + false_negatives == 0:
            return 1.0  # Only true negatives (so all predictions are correct)

        return 2 * true_positives / (2 * true_positives + false_positives + false_negatives)

    @staticmethod
    def precision(predicted: Column | TabularDataset, expected: Column | TabularDataset, positive_class: Any) -> float:
        """
        Compute the precision on the given data.

        The precision is the proportion of positive predictions that were correct. The **higher** the precision, the
        better the classifier. Results range from 0.0 to 1.0.

        Parameters
        ----------
        predicted:
            The predicted target values produced by the classifier.
        expected:
            The expected target values.
        positive_class:
            The class to be considered positive. All other classes are considered negative.

        Returns
        -------
        precision:
            The calculated precision.
        """
        expected = _extract_target(expected)
        predicted = _extract_target(predicted)
        _check_equal_length(predicted, expected)

        true_positives = (expected._series.eq(positive_class) & predicted._series.eq(positive_class)).sum()
        predicted_positives = predicted._series.eq(positive_class).sum()

        if predicted_positives == 0:
            return 1.0  # All positive predictions were correct (since there are none)

        return true_positives / predicted_positives

    @staticmethod
    def recall(predicted: Column | TabularDataset, expected: Column | TabularDataset, positive_class: Any) -> float:
        """
        Compute the recall on the given data.

        The recall is the proportion of actual positives that were predicted correctly. The **higher** the recall, the
        better the classifier. Results range from 0.0 to 1.0.

        Parameters
        ----------
        predicted:
            The predicted target values produced by the classifier.
        expected:
            The expected target values.
        positive_class:
            The class to be considered positive. All other classes are considered negative.

        Returns
        -------
        recall:
            The calculated recall.
        """
        expected = _extract_target(expected)
        predicted = _extract_target(predicted)
        _check_equal_length(predicted, expected)

        true_positives = (expected._series.eq(positive_class) & predicted._series.eq(positive_class)).sum()
        actual_positives = expected._series.eq(positive_class).sum()

        if actual_positives == 0:
            return 1.0  # All actual positives were predicted correctly (since there are none)

        return true_positives / actual_positives


def _extract_target(column_or_dataset: Column | TabularDataset) -> Column:
    """Extract the target column from the given column or dataset."""
    if isinstance(column_or_dataset, TabularDataset):
        return column_or_dataset.target
    else:
        return column_or_dataset


# TODO: collect validation in one place?
def _check_equal_length(column1: Column, column2: Column) -> None:
    """
    Check if the columns have the same length and raise an error if they do not.

    Parameters
    ----------
    column1:
        The first column.
    column2:
        The second column.

    Raises
    ------
    ValueError
        If the columns have different lengths.
    """
    if column1.row_count != column2.row_count:
        ColumnLengthMismatchError("")  # TODO: pass list of columns to exception, let it handle the formatting
